"""
Generic energy system definition
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import Union

from hub.city_model_structure.energy_systems.generic_distribution_system import GenericDistributionSystem
from hub.city_model_structure.energy_systems.generic_emission_system import GenericEmissionSystem
from hub.city_model_structure.energy_systems.generic_generation_system import GenericGenerationSystem


class GenericEnergySystem:
  """
  GenericEnergySystem class
  """
  def __init__(self):
    self._name = None
    self._demand_types = None
    self._generation_system = None
    self._distribution_system = None
    self._emission_system = None
    self._connected_city_objects = None

  @property
  def name(self):
    """
    Get energy system name
    :return: str
    """
    return self._name

  @name.setter
  def name(self, value):
    """
    Set energy system name
    :param value:
    """
    self._name = value

  @property
  def demand_types(self):
    """
    Get demand able to cover from [Heating, Cooling, Domestic Hot Water, Electricity]
    :return: [string]
    """
    return self._demand_types

  @demand_types.setter
  def demand_types(self, value):
    """
    Set demand able to cover from [Heating, Cooling, Domestic Hot Water, Electricity]
    :param value: [string]
    """
    self._demand_types = value

  @property
  def generation_system(self) -> GenericGenerationSystem:
    """
    Get generation system
    :return: GenerationSystem
    """
    return self._generation_system

  @generation_system.setter
  def generation_system(self, value):
    """
    Set generation system
    :return: GenerationSystem
    """
    self._generation_system = value

  @property
  def distribution_system(self) -> Union[None, GenericDistributionSystem]:
    """
    Get distribution system
    :return: DistributionSystem
    """
    return self._distribution_system

  @distribution_system.setter
  def distribution_system(self, value):
    """
    Set distribution system
    :param value: DistributionSystem
    """
    self._distribution_system = value

  @property
  def emission_system(self) -> Union[None, GenericEmissionSystem]:
    """
    Get emission system
    :return: EmissionSystem
    """
    return self._emission_system

  @emission_system.setter
  def emission_system(self, value):
    """
    Set emission system
    :param value: EmissionSystem
    """
    self._emission_system = value



"""
Capital costs included in the catalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List
from hub.catalog_factories.data_models.cost.chapter import Chapter


class CapitalCost:
  """
  Capital cost class
  """
  def __init__(self, general_chapters, design_allowance, overhead_and_profit):
    self._general_chapters = general_chapters
    self._design_allowance = design_allowance
    self._overhead_and_profit = overhead_and_profit

  @property
  def general_chapters(self) -> List[Chapter]:
    """
    Get general chapters in capital costs
    :return: [Chapter]
    """
    return self._general_chapters

  @property
  def design_allowance(self):
    """
    Get design allowance in percentage (-)
    :return: float
    """
    return self._design_allowance

  @property
  def overhead_and_profit(self):
    """
    Get overhead profit in percentage (-)
    :return: float
    """
    return self._overhead_and_profit

  def chapter(self, name) -> Chapter:
    """
    Get specific chapter by name
    :return: Chapter
    """
    for chapter in self.general_chapters:
      if chapter.chapter_type == name:
        return chapter
    raise KeyError(f'Chapter name {name} not found')

  def to_dictionary(self):
    """Class content to dictionary"""
    _chapters = []
    for _chapter in self.general_chapters:
      _chapters.append(_chapter.to_dictionary())
    content = {'Capital cost': {'design allowance': self.design_allowance,
                                'overhead and profit': self.overhead_and_profit,
                                'chapters': _chapters
                                }
               }

    return content



"""
ThermalControl module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from typing import Union, List
from hub.city_model_structure.attributes.schedule import Schedule


class ThermalControl:
  """
  ThermalControl class
  """
  def __init__(self):
    self._mean_heating_set_point = None
    self._heating_set_back = None
    self._mean_cooling_set_point = None
    self._hvac_availability_schedules = None
    self._heating_set_point_schedules = None
    self._cooling_set_point_schedules = None

  @staticmethod
  def _maximum_value(schedules):
    maximum = -1000
    for schedule in schedules:
      for value in schedule.values:
        if value > maximum:
          maximum = value
    return maximum

  @staticmethod
  def _minimum_value(schedules):
    minimum = 1000
    for schedule in schedules:
      for value in schedule.values:
        if value < minimum:
          minimum = value
    return minimum

  @property
  def mean_heating_set_point(self) -> Union[None, float]:
    """
    Get heating set point defined for a thermal zone in Celsius
    :return: None or float
    """
    if self._mean_heating_set_point is None:
      if self.heating_set_point_schedules is not None:
        self._mean_heating_set_point = self._maximum_value(self.heating_set_point_schedules)
    return self._mean_heating_set_point

  @mean_heating_set_point.setter
  def mean_heating_set_point(self, value):
    """
    Set heating set point defined for a thermal zone in Celsius
    :param value: float
    """
    self._mean_heating_set_point = float(value)

  @property
  def heating_set_back(self) -> Union[None, float]:
    """
    Get heating set back defined for a thermal zone in Celsius
    :return: None or float
    """
    if self._heating_set_back is None:
      if self.heating_set_point_schedules is not None:
        self._heating_set_back = self._minimum_value(self.heating_set_point_schedules)
    return self._heating_set_back

  @heating_set_back.setter
  def heating_set_back(self, value):
    """
    Set heating set back defined for a thermal zone in Celsius
    :param value: float
    """
    if value is not None:
      self._heating_set_back = float(value)

  @property
  def mean_cooling_set_point(self) -> Union[None, float]:
    """
    Get cooling set point defined for a thermal zone in Celsius
    :return: None or float
    """
    if self._mean_cooling_set_point is None:
      if self.cooling_set_point_schedules is not None:
        self._mean_cooling_set_point = self._minimum_value(self.cooling_set_point_schedules)
    return self._mean_cooling_set_point

  @mean_cooling_set_point.setter
  def mean_cooling_set_point(self, value):
    """
    Set cooling set point defined for a thermal zone in Celsius
    :param value: float
    """
    self._mean_cooling_set_point = float(value)

  @property
  def hvac_availability_schedules(self) -> Union[None, List[Schedule]]:
    """
    Get the availability of the conditioning system defined for a thermal zone
    dataType = on/off
    :return: None or [Schedule]
    """
    return self._hvac_availability_schedules

  @hvac_availability_schedules.setter
  def hvac_availability_schedules(self, value):
    """
    Set the availability of the conditioning system defined for a thermal zone
    dataType = on/off
    :param value: [Schedule]
    """
    self._hvac_availability_schedules = value

  @property
  def heating_set_point_schedules(self) -> Union[None, List[Schedule]]:
    """
    Get heating set point schedule defined for a thermal zone in Celsius
    dataType = temperature
    :return: None or [Schedule]
    """
    return self._heating_set_point_schedules

  @heating_set_point_schedules.setter
  def heating_set_point_schedules(self, value):
    """
    Set heating set point schedule defined for a thermal zone in Celsius
    dataType = temperature
    :param value: [Schedule]
    """
    self._heating_set_point_schedules = value

  @property
  def cooling_set_point_schedules(self) -> Union[None, List[Schedule]]:
    """
    Get cooling set point schedule defined for a thermal zone in Celsius
    dataType = temperature
    :return: None or [Schedule]
    """
    return self._cooling_set_point_schedules

  @cooling_set_point_schedules.setter
  def cooling_set_point_schedules(self, value):
    """
    Set cooling set point schedule defined for a thermal zone in Celsius
    dataType = temperature
    :param value: [Schedule]
    """
    self._cooling_set_point_schedules = value



"""
WeatherFactory retrieve the specific weather module for the given source format
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path

from hub.city_model_structure.city import City
from hub.helpers.utils import validate_import_export_type
from hub.imports.weather.epw_weather_parameters import EpwWeatherParameters


class WeatherFactory:
  """
  WeatherFactory class
  """

  def __init__(self, handler, city: City):
    self._handler = '_' + handler.lower().replace(' ', '_')
    validate_import_export_type(WeatherFactory, handler)
    self._city = city

  def _epw(self):
    """
    Enrich the city with energy plus weather file
    """
    # EnergyPlus Weather
    # to download files: https://energyplus.net/weather
    # description of the format: https://energyplus.net/sites/default/files/pdfs_v8.3.0/AuxiliaryPrograms.pdf
    return EpwWeatherParameters(self._city)

  def enrich(self):
    """
    Enrich the city given to the class using the given weather handler
    :return: None
    """
    getattr(self, self._handler, lambda: None)()



"""
City repository with database CRUD operations
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
import datetime
import logging

from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from hub.city_model_structure.city import City as CityHub
from hub.persistence.repository import Repository
from hub.persistence.models import City as Model
from hub.persistence.models import CityObject
from hub.version import __version__


class City(Repository):
  """
  City repository
  """
  _instance = None

  def __init__(self, db_name: str, dotenv_path: str, app_env: str):
    super().__init__(db_name, dotenv_path, app_env)

  def __new__(cls, db_name, dotenv_path, app_env):
    """
    Implemented for a singleton pattern
    """
    if cls._instance is None:
      cls._instance = super(City, cls).__new__(cls)
    return cls._instance

  def insert(self, city: CityHub, pickle_path, application_id, user_id: int):
    """
    Inserts a city
    :param city: The complete city instance
    :param pickle_path: Path to the pickle
    :param application_id: Application id owning the instance
    :param user_id: User id owning the instance
    :return: Identity id
    """
    city.save_compressed(pickle_path)
    try:
      db_city = Model(
        pickle_path,
        city.name,
        city.level_of_detail.geometry,
        'None' if city.climate_file is None else str(city.climate_file),
        application_id,
        user_id,
        __version__)

      self.session.add(db_city)
      self.session.flush()
      self.session.commit()
      for building in city.buildings:
        db_city_object = CityObject(db_city.id,
                                    building)
        self.session.add(db_city_object)
        self.session.flush()
      self.session.commit()
      self.session.refresh(db_city)
      return db_city.id
    except SQLAlchemyError as err:
      logging.error('An error occurred while creating a city %s', err)
      raise SQLAlchemyError from err

  def update(self, city_id: int, city: CityHub):
    """
    Updates a city name (other updates makes no sense)
    :param city_id: the id of the city to be updated
    :param city: the city object
    :return: None
    """
    try:
      now = datetime.datetime.utcnow()
      self.session.query(Model).filter(Model.id == city_id).update({'name': city.name, 'updated': now})
      self.session.commit()
    except SQLAlchemyError as err:
      logging.error('Error while updating city %s', err)
      raise SQLAlchemyError from err

  def delete(self, city_id: int):
    """
    Deletes a City with the id
    :param city_id: the city id
    :return: None
    """
    try:
      self.session.query(CityObject).filter(CityObject.city_id == city_id).delete()
      self.session.query(Model).filter(Model.id == city_id).delete()
      self.session.commit()
    except SQLAlchemyError as err:
      logging.error('Error while fetching city %s', err)
      raise SQLAlchemyError from err

  def get_by_user_id_application_id_and_name(self, user_id, application_id, city_name) -> Model:
    """
    Fetch city based on the user who created it
    :param user_id: the user id
    :param application_id: the application id
    :param city_name: the city name
    :return: ModelCity
    """
    try:
      result_set = self.session.execute(select(Model).where(Model.user_id == user_id,
                                                            Model.application_id == application_id,
                                                            Model.name == city_name
                                                            )).first()
      if result_set is not None:
        result_set = result_set[0]
      return result_set
    except SQLAlchemyError as err:
      logging.error('Error while fetching city by name %s', err)
      raise SQLAlchemyError from err

  def get_by_user_id_and_application_id(self, user_id, application_id) -> [Model]:
    """
    Fetch city based on the user who created it
    :param user_id: the user id
    :param application_id: the application id
    :return: ModelCity
    """
    try:
      result_set = self.session.execute(
        select(Model).where(Model.user_id == user_id, Model.application_id == application_id)
      )
      return [r[0] for r in result_set]
    except SQLAlchemyError as err:
      logging.error('Error while fetching city by name %s', err)
      raise SQLAlchemyError from err



"""
air_source_hp module defines an air source heat pump
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Peter Yefi peteryefi@gmail.com
"""

from typing import List
from hub.city_model_structure.energy_systems.heat_pump import HeatPump


class AirSourceHP(HeatPump):
  """
  AirSourceHP class
  """

  def __init__(self):
    super().__init__()
    self._cooling_capacity = None
    self._cooling_comp_power = None
    self._cooling_capacity_coff = None  # a coefficients for insel
    self._heating_capacity = None
    self._heating_comp_power = None
    self._heating_capacity_coff = None

  @property
  def cooling_capacity(self) -> List[float]:
    """
    Get cooling capacity in kW
    :return: [[float]]
    """
    return self._cooling_capacity

  @cooling_capacity.setter
  def cooling_capacity(self, value):
    """
    Set cooling capacity in kW
    :param value: [[float]]
    """
    if self._cooling_capacity is None:
      self._cooling_capacity = value

  @property
  def cooling_comp_power(self) -> List[float]:
    """
    Get cooling compressor power input in kW
    :return: [[float]]
    """
    return self._cooling_comp_power

  @cooling_comp_power.setter
  def cooling_comp_power(self, value):
    """
    Set the cooling compressor in kW
    :param value: [[float]]
    :return:
    """
    if self._cooling_comp_power is None:
      self._cooling_comp_power = value

  @property
  def cooling_capacity_coff(self) -> List[float]:
    """
    Get cooling capacity coefficients
    :return: [float]
    """
    return self._cooling_capacity_coff

  @cooling_capacity_coff.setter
  def cooling_capacity_coff(self, value):
    """
    Set the value for cooling capacity coefficients
    :param value: [float]
    :return:
    """
    if self._cooling_capacity_coff is None:
      self._cooling_capacity_coff = value

  @property
  def heating_capacity(self) -> List[float]:
    """
    Get heating capacity kW
    :return: [[float]]
    """
    return self._heating_capacity

  @heating_capacity.setter
  def heating_capacity(self, value):
    """
    Set the heating capacity in kW
    :param value: [[float]]
    :return:
    """
    if self._heating_capacity is None:
      self._heating_capacity = value

  @property
  def heating_comp_power(self) -> List[float]:
    """
    Get heating compressor power kW
    :return: [[float]]
    """
    return self._heating_comp_power

  @heating_comp_power.setter
  def heating_comp_power(self, value):
    """
    Set the heating compressor power in kW
    :param value: [[float]]
    :return:
    """
    if self._heating_comp_power is None:
      self._heating_comp_power = value

  @property
  def heating_capacity_coff(self) -> List[float]:
    """
    Get heating capacity coefficients
    :return: [float]
    """
    return self._heating_capacity_coff

  @heating_capacity_coff.setter
  def heating_capacity_coff(self, value):
    """
    Set the value for heating capacity coefficients
    :param value: [float]
    :return:
    """
    if self._heating_capacity_coff is None:
      self._heating_capacity_coff = value



"""
Traffic light module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Guille guille.gutierrezmorote@concordia.ca
"""

import ast
from typing import List, Union
from hub.city_model_structure.transport.phase import Phase
from hub.city_model_structure.transport.traffic_node import TrafficNode


class TrafficLight(TrafficNode):
  """
  Traffic light class
  """
  def __init__(self, name, coordinates, offset, phases=None, edges=None, right_on_red=False):
    super().__init__(name, coordinates, edges=edges, node_type='TrafficLight')
    if phases is None:
      phases = []
    self._right_on_red = right_on_red
    self._offset = offset
    self._phases = phases

  @property
  def right_on_red(self) -> Union[None, bool]:
    """
    Get if is possible to turn right when the traffic light is red
    :return: None or Boolean
    """
    return self._right_on_red

  @right_on_red.setter
  def right_on_red(self, value):
    """
    Get if is possible to turn right when the traffic light is red
    :param value: Boolean
    """
    if value is not None:
      self._right_on_red = ast.literal_eval(value)

  @property
  def offset(self) -> Union[None, int]:
    """
    Get program initial time offset
    :return: None or int
    """
    return self._offset

  @offset.setter
  def offset(self, value):
    """
    Set program initial time offset
    :param value: int
    """
    if value is not None:
      self._offset = int(value)

  @property
  def phases(self) -> List[Phase]:
    """
    Get traffic light logic phases
    :return: [Phase]
    """
    return self._phases

  @phases.setter
  def phases(self, value):
    """
    Set traffic light logic phases
    :param value: [Phase]
    """
    self._phases = value



"""
Phase module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import List, Union


class Phase:
  """
  Phase class
  """

  def __init__(self):
    self._duration = None
    self._state = None
    self._min_duration = None
    self._max_duration = None
    self._name = None
    self._next = None

  @property
  def duration(self) -> Union[None, int]:
    """
    Get phase duration in seconds
    :return: None or int
    """
    return self._duration

  @duration.setter
  def duration(self, value):
    """
    Set phase duration in seconds
    :param value: int
    """
    if value is not None:
      self._duration = int(value)

  @property
  def state(self) -> Union[None, List[str]]:
    """
    Get the list of signal states
    :return: None or [str]
    """
    return self._state

  @state.setter
  def state(self, value):
    """
    Set the list of signal states
    :param value: [str]
    """
    if value is not None:
      self._state = [str(i) for i in value]

  @property
  def min_duration(self) -> Union[None, int]:
    """
    Get phase minimum duration in seconds
    :return: None or int
    """
    if self._min_duration is None:
      self._min_duration = self._duration
    return self._min_duration

  @min_duration.setter
  def min_duration(self, value):
    """
    Set phase minimum duration in seconds
    :param value: int
    """
    if value is not None:
      self._min_duration = int(value)

  @property
  def max_duration(self) -> Union[None, int]:
    """
    Get phase maximum duration in seconds
    :return: None or int
    """
    if self._max_duration is None:
      self._max_duration = self._duration
    return self._max_duration

  @max_duration.setter
  def max_duration(self, value):
    """
    Set phase maximum duration in seconds
    :param value: int
    """
    if value is not None:
      self._max_duration = int(value)

  @property
  def name(self) -> Union[None, str]:
    """
    Get phase name
    :return: None or str
    """
    return self._name

  @name.setter
  def name(self, value):
    """
    Set phase name
    :param value: str
    """
    if value is not None:
      self._name = str(value)

  @property
  def next(self) -> Union[None, List[int]]:
    """
    Get the next phase in the cycle after the current.
    This is useful when adding extra transition phases to a traffic light plan which are not part of every cycle.
    Traffic lights of type 'actuated' can make use of a list of indices for selecting among alternative
    successor phases.
    :return: None or [int]
    """
    return self._next

  @next.setter
  def next(self, value):
    """
    Get the next phase in the cycle after the current.
    This is useful when adding extra transition phases to a traffic light plan which are not part of every cycle.
    Traffic lights of type 'actuated' can make use of a list of indices for selecting among alternative
    successor phases.
    :param value: [int]
    """
    if value is not None:
      self._next = [int(i) for i in value]



"""
Test db factory
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
import glob
import json
import logging
import os
import subprocess
import unittest
from unittest import TestCase
from pathlib import Path
import sqlalchemy.exc

from hub.imports.geometry_factory import GeometryFactory
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.usage_factory import UsageFactory
from hub.imports.results_factory import ResultFactory
from hub.imports.weather_factory import WeatherFactory
from hub.imports.energy_systems_factory import EnergySystemsFactory
from hub.exports.energy_building_exports_factory import EnergyBuildingsExportsFactory
from hub.exports.exports_factory import ExportsFactory
from hub.persistence.db_control import DBControl
from hub.persistence.repository import Repository
from sqlalchemy import create_engine
from hub.persistence.models import City, Application, CityObject, SimulationResults
from hub.persistence.models import User, UserRoles
from hub.helpers.dictionaries import Dictionaries
from sqlalchemy.exc import ProgrammingError
import uuid
import hub.helpers.constants as cte
import distutils.spawn


class Control:
  _skip_test = False
  _skip_reason = 'PostgreSQL not properly installed in host machine'

  def __init__(self):
    """
    Test
    setup
    :return: None
    """
    self._skip_test = False
    # Create test database
    dotenv_path = Path("{}/.local/etc/hub/.env".format(os.path.expanduser('~'))).resolve()
    if not dotenv_path.exists():
      self._skip_test = True
      self._skip_reason = f'.env file missing at {dotenv_path}'
      return
    dotenv_path = str(dotenv_path)
    repository = Repository(db_name='hub_unittests', app_env='TEST', dotenv_path=dotenv_path)
    engine = create_engine(repository.configuration.connection_string)
    try:
      # delete test database if it exists
      connection = engine.connect()
      connection.close()
    except ProgrammingError:
      logging.info('Database does not exist. Nothing to delete')
    except sqlalchemy.exc.OperationalError as operational_error:
      self._skip_test = True
      self._skip_reason = f'{operational_error}'
      return

    Application.__table__.create(bind=repository.engine, checkfirst=True)
    User.__table__.create(bind=repository.engine, checkfirst=True)
    City.__table__.create(bind=repository.engine, checkfirst=True)
    CityObject.__table__.create(bind=repository.engine, checkfirst=True)
    SimulationResults.__table__.create(bind=repository.engine, checkfirst=True)

    city_file = "tests_data/FZK_Haus_LoD_2.gml"
    output_path = Path('tests_outputs/').resolve()
    self._city = GeometryFactory('citygml',
                                 city_file,
                                 function_to_hub=Dictionaries().alkis_function_to_hub_function).city
    ConstructionFactory('nrcan', self._city).enrich()
    UsageFactory('nrcan', self._city).enrich()
    WeatherFactory('epw', self._city).enrich()
    ExportsFactory('sra', self._city, output_path).export()
    sra_file = str((output_path / f'{self._city.name}_sra.xml').resolve())
    subprocess.run([self.sra, sra_file], stdout=subprocess.DEVNULL)
    ResultFactory('sra', self._city, output_path).enrich()


    for building in self._city.buildings:
      building.energy_systems_archetype_name = 'system 1 gas pv'
    EnergySystemsFactory('montreal_custom', self._city).enrich()
    EnergyBuildingsExportsFactory('insel_monthly_energy_balance', self._city, output_path).export()
    _insel_files = glob.glob(f'{output_path}/*.insel')
    for insel_file in _insel_files:
      subprocess.run([self.insel, str(insel_file)], stdout=subprocess.DEVNULL)
    ResultFactory('insel_monthly_energy_balance', self._city, output_path).enrich()

    self._database = DBControl(
      db_name=repository.configuration.db_name,
      app_env='TEST',
      dotenv_path=dotenv_path)

    self._application_uuid = str(uuid.uuid4())
    self._application_id = self._database.persist_application('test', 'test application', self.application_uuid)
    self._user_id = self._database.create_user('Admin', self._application_id, 'Admin@123', UserRoles.Admin)
    self._pickle_path = 'tests_data/pickle_path.bz2'

  @property
  def database(self):
    return self._database

  @property
  def application_uuid(self):
    return self._application_uuid

  @property
  def application_id(self):
    return self._application_id

  @property
  def user_id(self):
    return self._user_id

  @property
  def skip_test(self):
    return self._skip_test

  @property
  def insel(self):
    return distutils.spawn.find_executable('insel')

  @property
  def sra(self):
    return distutils.spawn.find_executable('sra')

  @property
  def skip_insel_test(self):
    return self.insel is None

  @property
  def skip_reason(self):
    return self._skip_reason

  @property
  def message(self):
    return self._skip_reason

  @property
  def city(self):
    return self._city

  @property
  def pickle_path(self):
    return self._pickle_path


control = Control()


class TestDBFactory(TestCase):
  """
TestDBFactory
"""

  @unittest.skipIf(control.skip_test, control.skip_reason)
  def test_save_city(self):
    control.city.name = "Montreal"
    city_id = control.database.persist_city(
      control.city,
      control.pickle_path,
      control.application_id,
      control.user_id)
    control.database.delete_city(city_id)
    os.unlink(control.pickle_path)

  @unittest.skipIf(control.skip_test, control.skip_reason)
  def test_get_update_city(self):
    city_id = control.database.persist_city(control.city,
                                            control.pickle_path,
                                            control.application_id,
                                            control.user_id)
    control.city.name = "Ottawa"
    control.database.update_city(city_id, control.city)
    cities = control.database.cities_by_user_and_application(
      control.user_id,
      control.application_id)
    for updated_city in cities:
      if updated_city.id == city_id:
        self.assertEqual(updated_city.name, control.city.name)
        break
    control.database.delete_city(city_id)

  @unittest.skipIf(control.skip_test, control.skip_reason)
  @unittest.skipIf(control.skip_insel_test, 'insel is not installed')
  def test_save_results(self):
    city_id = control.database.persist_city(control.city,
                                            control.pickle_path,
                                            control.application_id,
                                            control.user_id)
    city_objects_id = []
    for building in control.city.buildings:
      _building = control.database.building_info(building.name, city_id)
      if cte.MONTH not in building.cooling_demand:
        print(f'building {building.name} not calculated')
        continue
      monthly_cooling_peak_load = building.cooling_peak_load[cte.MONTH]
      yearly_cooling_peak_load = building.cooling_peak_load[cte.YEAR]
      monthly_heating_peak_load = building.heating_peak_load[cte.MONTH]
      yearly_heating_peak_load = building.heating_peak_load[cte.YEAR]
      monthly_cooling_demand = building.cooling_demand[cte.MONTH][cte.INSEL_MEB]
      yearly_cooling_demand = building.cooling_demand[cte.YEAR][cte.INSEL_MEB]
      monthly_heating_demand = building.heating_demand[cte.MONTH][cte.INSEL_MEB]
      yearly_heating_demand = building.heating_demand[cte.YEAR][cte.INSEL_MEB]
      monthly_lighting_electrical_demand = building.lighting_electrical_demand[cte.MONTH][cte.INSEL_MEB]
      yearly_lighting_electrical_demand = building.lighting_electrical_demand[cte.YEAR][cte.INSEL_MEB]
      monthly_appliances_electrical_demand = building.appliances_electrical_demand[cte.MONTH][cte.INSEL_MEB]
      yearly_appliances_electrical_demand = building.appliances_electrical_demand[cte.YEAR][cte.INSEL_MEB]
      monthly_domestic_hot_water_heat_demand = building.domestic_hot_water_heat_demand[cte.MONTH][cte.INSEL_MEB]
      yearly_domestic_hot_water_heat_demand = building.domestic_hot_water_heat_demand[cte.YEAR][cte.INSEL_MEB]
      monthly_heating_consumption = building.heating_consumption[cte.MONTH]
      yearly_heating_consumption = building.heating_consumption[cte.YEAR]
      monthly_cooling_consumption = building.cooling_consumption[cte.MONTH]
      yearly_cooling_consumption = building.cooling_consumption[cte.YEAR]
      monthly_domestic_hot_water_consumption = building.domestic_hot_water_consumption[cte.MONTH]
      yearly_domestic_hot_water_consumption = building._domestic_hot_water_consumption[cte.YEAR]
      monthly_distribution_systems_electrical_consumption = building.distribution_systems_electrical_consumption[cte.MONTH]
      yearly_distribution_systems_electrical_consumption = building.distribution_systems_electrical_consumption[cte.YEAR]
      monthly_on_site_electrical_production = building.onsite_electrical_production[cte.MONTH]
      yearly_on_site_electrical_production = building.onsite_electrical_production[cte.YEAR]
      results = json.dumps({cte.INSEL_MEB: [
          {'monthly_cooling_peak_load': monthly_cooling_peak_load},
          {'yearly_cooling_peak_load': yearly_cooling_peak_load},
          {'monthly_heating_peak_load': monthly_heating_peak_load},
          {'yearly_heating_peak_load': yearly_heating_peak_load},
          {'monthly_cooling_demand': monthly_cooling_demand.tolist()},
          {'yearly_cooling_demand': yearly_cooling_demand.tolist()},
          {'monthly_heating_demand': monthly_heating_demand.tolist()},
          {'yearly_heating_demand': yearly_heating_demand.tolist()},
          {'monthly_lighting_electrical_demand': monthly_lighting_electrical_demand.tolist()},
          {'yearly_lighting_electrical_demand': yearly_lighting_electrical_demand.tolist()},
          {'monthly_appliances_electrical_demand': monthly_appliances_electrical_demand.tolist()},
          {'yearly_appliances_electrical_demand': yearly_appliances_electrical_demand.tolist()},
          {'monthly_domestic_hot_water_heat_demand': monthly_domestic_hot_water_heat_demand.tolist()},
          {'yearly_domestic_hot_water_heat_demand': yearly_domestic_hot_water_heat_demand.tolist()},
          {'monthly_heating_consumption': monthly_heating_consumption},
          {'yearly_heating_consumption': yearly_heating_consumption},
          {'monthly_cooling_consumption': monthly_cooling_consumption},
          {'yearly_cooling_consumption': yearly_cooling_consumption},
          {'monthly_domestic_hot_water_consumption': monthly_domestic_hot_water_consumption},
          {'yearly_domestic_hot_water_consumption': yearly_domestic_hot_water_consumption},
          {'monthly_distribution_systems_electrical_consumption': monthly_distribution_systems_electrical_consumption},
          {'yearly_distribution_systems_electrical_consumption': yearly_distribution_systems_electrical_consumption},
          {'monthly_on_site_electrical_production': monthly_on_site_electrical_production},
          {'yearly_on_site_electrical_production': yearly_on_site_electrical_production}
        ]})

      db_building_id = _building.id
      city_objects_id.append(db_building_id)
      control.database.add_simulation_results(
        cte.INSEL_MEB,
        results, city_object_id=db_building_id)
    self.assertEqual(1, len(city_objects_id), 'wrong number of results')
    self.assertIsNotNone(city_objects_id[0], 'city_object_id is None')
    for _id in city_objects_id:
      control.database.delete_results_by_name('insel meb', city_object_id=_id)
    control.database.delete_city(city_id)

  @classmethod
  @unittest.skipIf(control.skip_test, control.skip_reason)
  def tearDownClass(cls):
    control.database.delete_application(control.application_uuid)
    control.database.delete_user(control.user_id)



"""
Node module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

import uuid
from typing import List, TypeVar
from hub.city_model_structure.attributes.time_series import TimeSeries
Edge = TypeVar('Edge')


class Node:
  """
  Generic node class to be used as base for the network nodes
  """
  def __init__(self, name, edges=None):
    if edges is None:
      edges = []
    self._name = name
    self._id = None
    self._edges = edges
    self._time_series = None

  @property
  def name(self):
    """
    Get node name
    :return: str
    """
    return self._name

  @property
  def id(self):
    """
    Get node id, a universally unique identifier randomly generated
    :return: str
    """
    if self._id is None:
      self._id = uuid.uuid4()
    return self._id

  @property
  def edges(self) -> List[Edge]:
    """
    Get edges delimited by the node
    :return: [Edge]
    """
    return self._edges

  @property
  def time_series(self) -> TimeSeries:
    """
    Add explanation here
    :return: add type of variable here
    """
    return self._time_series



"""
WaterToWaterHPExport exports water to water values after executing insel.
Multiple files are generated for the export
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
from hub.exports.energy_systems.heat_pump_export import HeatPumpExport
from typing import List, Dict, Union


class WaterToWaterHPExport(HeatPumpExport):
  """
  Exports heat pump values as multiple output files
  after executing insel
  """

  def __init__(self, base_path, city, output_path, sim_type, demand_path):
    """
    :param base_path: path to energy system files
    :param city: the city object
    :param output_path: the file to hold insel simulation results
    :param sim_type: the simulation type to run: 1 for series, 0 for parallel
    :param demand_path: path to hourly energy demand file
    """
    tmp_file = 'heat_pumps/w2w_series.txt' if sim_type == 0 else 'heat_pumps/w2w_parallel.txt'
    template_path = (base_path / tmp_file)
    water_temp = (base_path / 'heat_pumps/wt_hourly3.txt')
    super().__init__(base_path=base_path, city=city, output_path=output_path, template=template_path,
                     demand_path=demand_path, water_temp=water_temp)

  def _extract_model_coefficient(self, hp_model: str) -> Union[List, None]:
    """
    Extracts heat pump coefficient data for a specific
    model. e.g. ClimateMaster 156 kW, etc.
    :param hp_model: the model type
    :return:
    """
    for energy_system in self._city.energy_systems:
      if energy_system.water_to_water_hp.model == hp_model:
        return energy_system.water_to_water_hp.power_demand_coff
    return None

  def execute_insel(self, user_input, hp_model) -> Union[Dict, None]:
    """
    Runs insel and produces output files
    Runs insel and write the necessary files
    :param user_input: a dictionary containing the user
    values necessary to run insel
    :param hp_model: a string that indicates the heat
    pump model to be used e.g. 012, 015
    :return:
    """
    pow_demand_coefficient = self._extract_model_coefficient(hp_model)
    return super(WaterToWaterHPExport, self)._run_insel(user_input, pow_demand_coefficient, 'w2w.insel')



"""
Greenery in idf test
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
from pathlib import Path
from unittest import TestCase

import hub.helpers.constants as cte
from hub.city_model_structure.greenery.plant import Plant
from hub.city_model_structure.greenery.soil import Soil
from hub.city_model_structure.greenery.vegetation import Vegetation
from hub.exports.energy_building_exports_factory import EnergyBuildingsExportsFactory
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.usage_factory import UsageFactory
from hub.imports.weather_factory import WeatherFactory


class GreeneryInIdf(TestCase):
  """
  GreeneryInIdf TestCase 1
  """
  def test_greenery_in_idf(self):
    city_file = "tests_data/one_building_in_kelowna.gml"
    output_path = Path('tests_outputs/').resolve()

    city = GeometryFactory('citygml', path=city_file).city
    for building in city.buildings:
      building.year_of_construction = 2006
    ConstructionFactory('nrel', city).enrich()
    UsageFactory('comnet', city).enrich()
    WeatherFactory('epw', city).enrich()
    vegetation_name = 'BaseEco'
    soil_thickness = 0.18
    soil_name = 'EcoRoofSoil'
    roughness = 'MediumSmooth'
    dry_conductivity = 0.4
    dry_density = 641
    dry_specific_heat = 1100
    thermal_absorptance = 0.95
    solar_absorptance = 0.8
    visible_absorptance = 0.7
    saturation_volumetric_moisture_content = 0.4
    residual_volumetric_moisture_content = 0.01
    soil = Soil(soil_name, roughness, dry_conductivity, dry_density, dry_specific_heat, thermal_absorptance,
                solar_absorptance, visible_absorptance, saturation_volumetric_moisture_content,
                residual_volumetric_moisture_content)
    soil.initial_volumetric_moisture_content = 0.2
    plant_name = 'plant'
    height = 0.5
    leaf_area_index = 5
    leaf_reflectivity = 0.2
    leaf_emissivity = 0.95
    minimal_stomatal_resistance = 180
    co2_sequestration = 0
    grows_on_soils = [soil]
    plant = Plant(plant_name, height, leaf_area_index, leaf_reflectivity, leaf_emissivity, minimal_stomatal_resistance,
                  co2_sequestration, grows_on_soils)
    plant.percentage = 1
    plants = [plant]
    vegetation = Vegetation(vegetation_name, soil, soil_thickness, plants)
    for building in city.buildings:
      for surface in building.surfaces:
        if surface.type == cte.ROOF:
          surface.vegetation = vegetation

    _idf = EnergyBuildingsExportsFactory('idf', city, output_path).export()
    self.assertIsNotNone(_idf)
    city = GeometryFactory('citygml', path=city_file).city
    for building in city.buildings:
      building.year_of_construction = 2006
    ConstructionFactory('nrel', city).enrich()
    UsageFactory('comnet', city).enrich()
    WeatherFactory('epw', city).enrich()
    _idf = EnergyBuildingsExportsFactory('idf', city, output_path).export()
    self.assertIsNotNone(_idf)



"""
Usage catalog domestic hot water
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import Union, List

from hub.catalog_factories.data_models.usages.schedule import Schedule


class DomesticHotWater:
  """
  DomesticHotWater class
  """
  def __init__(self, density, peak_flow, service_temperature, schedules):
    self._density = density
    self._peak_flow = peak_flow
    self._service_temperature = service_temperature
    self._schedules = schedules

  @property
  def density(self) -> Union[None, float]:
    """
    Get domestic hot water load density in Watts per m2
    :return: None or float
    """
    return self._density

  @property
  def peak_flow(self) -> Union[None, float]:
    """
    Get domestic hot water peak_flow density in m3 per second and m2
    :return: None or float
    """
    return self._peak_flow

  @property
  def service_temperature(self) -> Union[None, float]:
    """
    Get service temperature in degrees Celsius
    :return: None or float
    """
    return self._service_temperature

  @property
  def schedules(self) -> Union[None, List[Schedule]]:
    """
    Get schedules
    dataType = fraction of loads
    :return: None or [Schedule]
    """
    return self._schedules

  def to_dictionary(self):
    """Class content to dictionary"""
    _schedules = []
    for _schedule in self.schedules:
      _schedules.append(_schedule.to_dictionary())
    content = {'Domestic hot water': {'density [W/m2]': self.density,
                                      'peak flow [m3/sm2]': self.peak_flow,
                                      'service temperature [Celsius]': self.service_temperature,
                                      'schedules': _schedules}
               }
    return content



"""
Building module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging
from typing import List, Union, TypeVar

import numpy as np
import hub.helpers.constants as cte
from hub.city_model_structure.attributes.polyhedron import Polyhedron
from hub.city_model_structure.building_demand.household import Household
from hub.city_model_structure.building_demand.internal_zone import InternalZone
from hub.city_model_structure.building_demand.thermal_zone import ThermalZone
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.city_object import CityObject
from hub.city_model_structure.energy_systems.energy_system import EnergySystem
from hub.helpers.peak_loads import PeakLoads

City = TypeVar('City')


class Building(CityObject):
  """
  Building(CityObject) class
  """
  def __init__(self, name, surfaces, year_of_construction, function, terrains=None, city=None):
    super().__init__(name, surfaces)
    self._city = city
    self._households = None
    self._basement_heated = None
    self._attic_heated = None
    self._terrains = terrains
    self._year_of_construction = year_of_construction
    self._function = function
    self._average_storey_height = None
    self._storeys_above_ground = None
    self._floor_area = None
    self._roof_type = None
    self._internal_zones = None
    self._thermal_zones = None
    self._shell = None
    self._aliases = []
    self._type = 'building'
    self._cold_water_temperature = {}
    self._heating_demand = {}
    self._cooling_demand = {}
    self._lighting_electrical_demand = {}
    self._appliances_electrical_demand = {}
    self._domestic_hot_water_heat_demand = {}
    self._heating_consumption = {}
    self._cooling_consumption = {}
    self._domestic_hot_water_consumption = {}
    self._distribution_systems_electrical_consumption = {}
    self._onsite_electrical_production = {}
    self._eave_height = None
    self._energy_systems = None
    self._systems_archetype_name = None
    self._grounds = []
    self._roofs = []
    self._walls = []
    self._internal_walls = []
    self._ground_walls = []
    self._attic_floors = []
    self._interior_slabs = []
    for surface_id, surface in enumerate(self.surfaces):
      self._min_x = min(self._min_x, surface.lower_corner[0])
      self._min_y = min(self._min_y, surface.lower_corner[1])
      self._min_z = min(self._min_z, surface.lower_corner[2])
      surface.id = surface_id
      if surface.type == cte.GROUND:
        self._grounds.append(surface)
      elif surface.type == cte.WALL:
        self._walls.append(surface)
      elif surface.type == cte.ROOF:
        self._roofs.append(surface)
      elif surface.type == cte.INTERIOR_WALL:
        self._internal_walls.append(surface)
      elif surface.type == cte.GROUND_WALL:
        self._ground_walls.append(surface)
      elif surface.type == cte.ATTIC_FLOOR:
        self._attic_floors.append(surface)
      elif surface.type == cte.INTERIOR_SLAB:
        self._interior_slabs.append(surface)
      else:
        logging.error(f'Building %s [%s] has an unexpected surface type %s.', self.name, self.aliases, surface.type)

  @property
  def shell(self) -> Polyhedron:
    """
    Get building's external polyhedron
    :return: [Polyhedron]
    """
    polygons = []
    for surface in self.surfaces:
      if surface.type is not cte.INTERIOR_WALL:
        polygons.append(surface.solid_polygon)
        if surface.holes_polygons is not None:
          for hole in surface.holes_polygons:
            polygons.append(hole)
    if self._shell is None:
      self._shell = Polyhedron(polygons)
    return self._shell

  @property
  def internal_zones(self) -> List[InternalZone]:
    """
    Get building internal zones
    For Lod up to 3, there is only one internal zone which corresponds to the building shell.
    In LoD 4 there can be more than one. In this case the definition of surfaces and floor area must be redefined.
    :return: [InternalZone]
    """
    if self._internal_zones is None:
      self._internal_zones = [InternalZone(self.surfaces, self.floor_area)]
    return self._internal_zones

  @property
  def thermal_zones(self) -> Union[None, List[ThermalZone]]:
    """
    Get building thermal zones
    For Lod up to 3, there can be more than one thermal zone per internal zone.
    In LoD 4, there can be more than one internal zone, and therefore, only one thermal zone per internal zone
    :return: [ThermalZone]
    """
    if self._thermal_zones is None:
      self._thermal_zones = []
      for internal_zone in self.internal_zones:
        if internal_zone.thermal_zones is None:
          self._thermal_zones = None
          return self._thermal_zones
        for thermal_zone in internal_zone.thermal_zones:
          self._thermal_zones.append(thermal_zone)
    return self._thermal_zones

  @property
  def grounds(self) -> List[Surface]:
    """
    Get building ground surfaces
    :return: [Surface]
    """
    return self._grounds

  @property
  def roofs(self) -> List[Surface]:
    """
    Get building roof surfaces
    :return: [Surface]
    """
    return self._roofs

  @property
  def walls(self) -> List[Surface]:
    """
    Get building wall surfaces
    :return: [Surface]
    """
    return self._walls

  @property
  def internal_walls(self) -> List[Surface]:
    """
    Get building internal wall surfaces
    :return: [Surface]
    """
    return self._internal_walls

  @property
  def terrains(self) -> Union[None, List[Surface]]:
    """
    Get city object terrain surfaces
    :return: [Surface]
    """
    return self._terrains

  @property
  def attic_heated(self) -> Union[None, int]:
    """
    Get if the city object attic is heated
    0: no attic in the building
    1: attic exists but is not heated
    2: attic exists and is heated
    :return: None or int
    """
    return self._attic_heated

  @attic_heated.setter
  def attic_heated(self, value):
    """
    Set if the city object attic is heated
    0: no attic in the building
    1: attic exists but is not heated
    2: attic exists and is heated
    :param value: int
    """
    if value is not None:
      self._attic_heated = int(value)

  @property
  def basement_heated(self) -> Union[None, int]:
    """
    Get if the city object basement is heated
    0: no basement in the building
    1: basement exists but is not heated
    2: basement exists and is heated
    :return: None or int
    """
    return self._basement_heated

  @basement_heated.setter
  def basement_heated(self, value):
    """
    Set if the city object basement is heated
    0: no basement in the building
    1: basement exists but is not heated
    2: basement exists and is heated
    :param value: int
    """
    if value is not None:
      self._basement_heated = int(value)

  @property
  def year_of_construction(self):
    """
    Get building year of construction
    :return: int
    """
    return self._year_of_construction

  @year_of_construction.setter
  def year_of_construction(self, value):
    """
    Set building year of construction
    :param value: int
    """
    if value is not None:
      self._year_of_construction = int(value)

  @property
  def function(self) -> Union[None, str]:
    """
    Get building function
    :return: None or str
    """
    return self._function

  @function.setter
  def function(self, value):
    """
    Set building function
    :param value: str
    """
    if value is not None:
      self._function = str(value)

  @property
  def average_storey_height(self) -> Union[None, float]:
    """
    Get building average storey height in meters
    :return: None or float
    """
    return self._average_storey_height

  @average_storey_height.setter
  def average_storey_height(self, value):
    """
    Set building average storey height in meters
    :param value: float
    """
    if value is not None:
      self._average_storey_height = float(value)

  @property
  def storeys_above_ground(self) -> Union[None, int]:
    """
    Get building storeys number above ground
    :return: None or int
    """
    if self._storeys_above_ground is None:
      if self.eave_height is not None and self.average_storey_height is not None:
        self._storeys_above_ground = int(self.eave_height / self.average_storey_height)
    return self._storeys_above_ground

  @storeys_above_ground.setter
  def storeys_above_ground(self, value):
    """
    Set building storeys number above ground
    :param value: int
    """
    if value is not None:
      self._storeys_above_ground = int(value)

  @property
  def cold_water_temperature(self) -> {float}:
    """
    Get cold water temperature in degrees Celsius
    :return: dict{DataFrame(float)}
    """
    return self._cold_water_temperature

  @cold_water_temperature.setter
  def cold_water_temperature(self, value):
    """
    Set cold water temperature in degrees Celsius
    :param value: dict{DataFrame(float)}
    """
    self._cold_water_temperature = value

  @property
  def heating_demand(self) -> dict:
    """
    Get heating demand in Wh
    :return: dict{DataFrame(float)}
    """
    return self._heating_demand

  @heating_demand.setter
  def heating_demand(self, value):
    """
    Set heating demand in Wh
    :param value: dict{DataFrame(float)}
    """
    self._heating_demand = value

  @property
  def cooling_demand(self) -> dict:
    """
    Get cooling demand in Wh
    :return: dict{DataFrame(float)}
    """
    return self._cooling_demand

  @cooling_demand.setter
  def cooling_demand(self, value):
    """
    Set cooling demand in Wh
    :param value: dict{DataFrame(float)}
    """
    self._cooling_demand = value

  @property
  def lighting_electrical_demand(self) -> dict:
    """
    Get lighting electrical demand in Wh
    :return: dict{DataFrame(float)}
    """
    return self._lighting_electrical_demand

  @lighting_electrical_demand.setter
  def lighting_electrical_demand(self, value):
    """
    Set lighting electrical demand in Wh
    :param value: dict{DataFrame(float)}
    """
    self._lighting_electrical_demand = value

  @property
  def appliances_electrical_demand(self) -> dict:
    """
    Get appliances electrical demand in Wh
    :return: dict{DataFrame(float)}
    """
    return self._appliances_electrical_demand

  @appliances_electrical_demand.setter
  def appliances_electrical_demand(self, value):
    """
    Set appliances electrical demand in Wh
    :param value: dict{DataFrame(float)}
    """
    self._appliances_electrical_demand = value

  @property
  def domestic_hot_water_heat_demand(self) -> dict:
    """
    Get domestic hot water heat demand in Wh
    :return: dict{DataFrame(float)}
    """
    return self._domestic_hot_water_heat_demand

  @domestic_hot_water_heat_demand.setter
  def domestic_hot_water_heat_demand(self, value):
    """
    Set domestic hot water heat demand in Wh
    :param value: dict{DataFrame(float)}
    """
    self._domestic_hot_water_heat_demand = value

  @property
  def heating_peak_load(self) -> Union[None, dict]:
    """
    Get heating peak load in W
    :return: dict{DataFrame(float)}
    """
    results = {}
    if cte.HOUR in self.heating_demand:
      monthly_values = PeakLoads().\
        peak_loads_from_hourly(self.heating_demand[cte.HOUR][next(iter(self.heating_demand[cte.HOUR]))])
    else:
      monthly_values = PeakLoads(self).heating_peak_loads_from_methodology
    if monthly_values is None:
      return None
    results[cte.MONTH] = monthly_values
    results[cte.YEAR] = [max(monthly_values)]
    return results

  @property
  def cooling_peak_load(self) -> Union[None, dict]:
    """
    Get cooling peak load in W
    :return: dict{DataFrame(float)}
    """
    results = {}
    if cte.HOUR in self.cooling_demand:
      monthly_values = PeakLoads().peak_loads_from_hourly(self.cooling_demand[cte.HOUR][next(iter(self.cooling_demand[cte.HOUR]))])
    else:
      monthly_values = PeakLoads(self).cooling_peak_loads_from_methodology
    if monthly_values is None:
      return None
    results[cte.MONTH] = monthly_values
    results[cte.YEAR] = [max(monthly_values)]
    return results

  @property
  def eave_height(self):
    """
    Get building eave height in meters
    :return: float
    """
    if self._eave_height is None:
      self._eave_height = 0
      for wall in self.walls:
        self._eave_height = max(self._eave_height, wall.upper_corner[2]) - self.simplified_polyhedron.min_z
    return self._eave_height

  @property
  def roof_type(self):
    """
    Get roof type for the building flat or pitch
    :return: str
    """
    if self._roof_type is None:
      self._roof_type = 'flat'
      for roof in self.roofs:
        grads = np.rad2deg(roof.inclination)
        if 355 > grads > 5:
          self._roof_type = 'pitch'
          break
    return self._roof_type

  @roof_type.setter
  def roof_type(self, value):
    """
    Set roof type for the building flat or pitch
    :return: str
    """
    self._roof_type = value

  @property
  def floor_area(self):
    """
    Get building floor area in square meters
    :return: float
    """
    if self._floor_area is None:
      self._floor_area = 0
      for surface in self.surfaces:
        if surface.type == 'Ground':
          self._floor_area += surface.perimeter_polygon.area
    return self._floor_area

  @property
  def households(self) -> List[Household]:
    """
    Get the list of households inside the building
    :return: List[Household]
    """
    return self._households

  @property
  def is_conditioned(self):
    """
    Get building heated flag
    :return: Boolean
    """
    if self.internal_zones is None:
      return False
    for internal_zone in self.internal_zones:
      if internal_zone.usages is not None:
        for usage in internal_zone.usages:
          if usage.thermal_control is not None:
            return True
    return False

  @property
  def aliases(self):
    """
    Get the alias name for the building
    :return: str
    """
    return self._aliases

  def add_alias(self, value):
    """
    Add a new alias for the building
    """
    self._aliases.append(value)
    if self.city is not None:
      self.city.add_building_alias(self, value)

  @property
  def city(self) -> City:
    """
    Get the city containing the building
    :return: City
    """
    return self._city

  @city.setter
  def city(self, value):
    """
    Set the city containing the building
    """
    self._city = value

  @property
  def usages_percentage(self):
    """
    Get the usages and percentages for the building
    """
    _usage = ''
    for internal_zone in self.internal_zones:
      if internal_zone.usages is None:
        continue
      for usage in internal_zone.usages:
        _usage = f'{_usage}{usage.name}_{usage.percentage} '
    return _usage.rstrip()

  @property
  def energy_systems(self) -> Union[None, List[EnergySystem]]:
    """
    Get list of energy systems installed to cover the building demands
    :return: [EnergySystem]
    """
    return self._energy_systems

  @energy_systems.setter
  def energy_systems(self, value):
    """
    Set list of energy systems installed to cover the building demands
    :param value: [EnergySystem]
    """
    self._energy_systems = value

  @property
  def energy_systems_archetype_name(self):
    """
    Get energy systems archetype name
    :return: str
    """
    return self._systems_archetype_name

  @energy_systems_archetype_name.setter
  def energy_systems_archetype_name(self, value):
    """
    Set energy systems archetype name
    :param value: str
    """
    self._systems_archetype_name = value

  @property
  def heating_consumption(self):
    """
    Get energy consumption for heating according to the heating system installed in Wh
    return: dict
    """
    if len(self._heating_consumption) == 0:
      for heating_demand_key in self.heating_demand:
        demand = self.heating_demand[heating_demand_key][cte.INSEL_MEB]
        consumption_type = cte.HEATING
        final_energy_consumed = self._calculate_consumption(consumption_type, demand)
        if final_energy_consumed is None:
          continue
        self._heating_consumption[heating_demand_key] = final_energy_consumed
    return self._heating_consumption

  @property
  def cooling_consumption(self):
    """
    Get energy consumption for cooling according to the cooling system installed in Wh
    return: dict
    """
    if len(self._cooling_consumption) == 0:
      for cooling_demand_key in self.cooling_demand:
        demand = self.cooling_demand[cooling_demand_key][cte.INSEL_MEB]
        consumption_type = cte.COOLING
        final_energy_consumed = self._calculate_consumption(consumption_type, demand)
        if final_energy_consumed is None:
          continue
        self._cooling_consumption[cooling_demand_key] = final_energy_consumed
    return self._cooling_consumption

  @property
  def domestic_hot_water_consumption(self):
    """
    Get energy consumption for domestic according to the domestic hot water system installed in Wh
    return: dict
    """
    if len(self._domestic_hot_water_consumption) == 0:
      for domestic_hot_water_demand_key in self.domestic_hot_water_heat_demand:
        demand = self.domestic_hot_water_heat_demand[domestic_hot_water_demand_key][cte.INSEL_MEB]
        consumption_type = cte.DOMESTIC_HOT_WATER
        final_energy_consumed = self._calculate_consumption(consumption_type, demand)
        if final_energy_consumed is None:
          continue
        self._domestic_hot_water_consumption[domestic_hot_water_demand_key] = final_energy_consumed
    return self._domestic_hot_water_consumption

  def _calculate_working_hours(self):
    _working_hours = {}
    for internal_zone in self.internal_zones:
      for thermal_zone in internal_zone.thermal_zones:
        _working_hours_per_thermal_zone = {}
        for schedule in thermal_zone.thermal_control.hvac_availability_schedules:
          _working_hours_per_schedule = [0] * len(schedule.values)
          for i, value in enumerate(schedule.values):
            if value > 0:
              _working_hours_per_schedule[i] = 1
          for day_type in schedule.day_types:
            _working_hours_per_thermal_zone[day_type] = _working_hours_per_schedule
        if len(_working_hours) == 0:
          _working_hours = _working_hours_per_thermal_zone
        else:
          for key, item in _working_hours.items():
            saved_values = _working_hours_per_thermal_zone[key]
            for i, value in enumerate(item):
              _working_hours[key][i] = max(_working_hours[key][i], saved_values[i])

    _total_hours = 0
    for key in _working_hours:
      hours = sum(_working_hours[key])
      _total_hours += hours * cte.DAYS_A_YEAR[key]
    return _total_hours

  @property
  def distribution_systems_electrical_consumption(self):
    """
    Get total electricity consumption for distribution and emission systems in Wh
    return: dict
    """
    if len(self._distribution_systems_electrical_consumption) != 0:
      return self._distribution_systems_electrical_consumption
    _peak_load = self.heating_peak_load[cte.YEAR][0]
    _peak_load_type = cte.HEATING
    if _peak_load < self.cooling_peak_load[cte.YEAR][0]:
      _peak_load = self.cooling_peak_load[cte.YEAR][0]
      _peak_load_type = cte.COOLING

    _working_hours = self._calculate_working_hours()
    _consumption_fix_flow = 0
    if self.energy_systems is None:
      return self._distribution_systems_electrical_consumption
    for energy_system in self.energy_systems:
      emission_system = energy_system.emission_system.generic_emission_system
      parasitic_energy_consumption = 0
      if emission_system is not None:
        parasitic_energy_consumption = emission_system.parasitic_energy_consumption
      distribution_system = energy_system.distribution_system.generic_distribution_system
      consumption_variable_flow = distribution_system.distribution_consumption_variable_flow
      for demand_type in energy_system.demand_types:
        if demand_type.lower() == cte.HEATING.lower():
          if _peak_load_type == cte.HEATING.lower():
            _consumption_fix_flow = distribution_system.distribution_consumption_fix_flow
          for heating_demand_key in self.heating_demand:
            _consumption = [0]*len(self.heating_demand[heating_demand_key][cte.INSEL_MEB])
            _demand = self.heating_demand[heating_demand_key][cte.INSEL_MEB]
            for i, _ in enumerate(_consumption):
              _consumption[i] += (parasitic_energy_consumption + consumption_variable_flow) * _demand[i]
            self._distribution_systems_electrical_consumption[heating_demand_key] = _consumption
        if demand_type.lower() == cte.COOLING.lower():
          if _peak_load_type == cte.COOLING.lower():
            _consumption_fix_flow = distribution_system.distribution_consumption_fix_flow
          for demand_key in self.cooling_demand:
            _consumption = self._distribution_systems_electrical_consumption[demand_key]
            _demand = self.cooling_demand[demand_key][cte.INSEL_MEB]
            for i, _ in enumerate(_consumption):
              _consumption[i] += (parasitic_energy_consumption + consumption_variable_flow) * _demand[i]
            self._distribution_systems_electrical_consumption[demand_key] = _consumption

      for key, item in self._distribution_systems_electrical_consumption.items():
        for i in range(0, len(item)):
          self._distribution_systems_electrical_consumption[key][i] += _peak_load * _consumption_fix_flow \
                                                                       * _working_hours
    return self._distribution_systems_electrical_consumption

  def _calculate_consumption(self, consumption_type, demand):
    # todo: modify when COP depends on the hour
    coefficient_of_performance = 0
    if self.energy_systems is None:
      return None
    for energy_system in self.energy_systems:
      for demand_type in energy_system.demand_types:
        if demand_type.lower() == consumption_type.lower():
          if consumption_type in (cte.HEATING, cte.DOMESTIC_HOT_WATER):
            coefficient_of_performance = energy_system.generation_system.generic_generation_system.heat_efficiency
          elif consumption_type == cte.COOLING:
            coefficient_of_performance = energy_system.generation_system.generic_generation_system.cooling_efficiency
          elif consumption_type == cte.ELECTRICITY:
            coefficient_of_performance = \
              energy_system.generation_system.generic_generation_system.electricity_efficiency
    if coefficient_of_performance == 0:
      values = [0]*len(demand)
      final_energy_consumed = values
    else:
      final_energy_consumed = []
      for demand_value in demand:
        final_energy_consumed.append(demand_value / coefficient_of_performance)
    return final_energy_consumed

  @property
  def onsite_electrical_production(self):
    """
    Get total electricity produced onsite in Wh
    return: dict
    """

    # Add other systems whenever new ones appear
    orientation_losses_factor = {cte.MONTH: {'north': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                             'east': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                                             'south': [2.137931, 1.645503, 1.320946, 1.107817, 0.993213, 0.945175,
                                                       0.967949, 1.065534, 1.24183, 1.486486, 1.918033, 2.210526],
                                             'west': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]},
                                 cte.YEAR: {'north': [0],
                                            'east': [0],
                                            'south': [1.212544],
                                            'west': [0]}
                                 }
    if self.energy_systems is None:
      return self._onsite_electrical_production
    for energy_system in self.energy_systems:
      if energy_system.generation_system.generic_generation_system.type == cte.PHOTOVOLTAIC:
        _efficiency = energy_system.generation_system.generic_generation_system.electricity_efficiency
        self._onsite_electrical_production = {}
        for _key in self.roofs[0].global_irradiance.keys():
          _results = [0 for _ in range(0, len(self.roofs[0].global_irradiance[_key][cte.SRA]))]
          for surface in self.roofs:
            if _key in orientation_losses_factor:
              _results = [x + y * _efficiency * surface.perimeter_area
                          * surface.solar_collectors_area_reduction_factor * z
                          for x, y, z in zip(_results, surface.global_irradiance[_key][cte.SRA],
                                             orientation_losses_factor[_key]['south'])]
          self._onsite_electrical_production[_key] = _results
    return self._onsite_electrical_production



"""
Constant module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

# universal constants

KELVIN = 273.15
WATER_DENSITY = 1000  # kg/m3
WATER_HEAT_CAPACITY = 4182  # J/kgK

AIR_DENSITY = 1.293  # kg/m3
AIR_HEAT_CAPACITY = 1005.2  # J/kgK


# converters
HOUR_TO_MINUTES = 60
MINUTES_TO_SECONDS = 60
HOUR_TO_SECONDS = 3600
METERS_TO_FEET = 3.28084
BTU_H_TO_WATTS = 0.29307107
KILO_WATTS_HOUR_TO_JULES = 3600000
GALLONS_TO_QUBIC_METERS = 0.0037854117954011185

# time
SECOND = 'second'
MINUTE = 'minute'
HOUR = 'hour'
DAY = 'day'
WEEK = 'week'
MONTH = 'month'
YEAR = 'year'

# day types
MONDAY = 'monday'
TUESDAY = 'tuesday'
WEDNESDAY = 'wednesday'
THURSDAY = 'thursday'
FRIDAY = 'friday'
SATURDAY = 'saturday'
SUNDAY = 'sunday'
HOLIDAY = 'holiday'
WINTER_DESIGN_DAY = 'winter_design_day'
SUMMER_DESIGN_DAY = 'summer_design_day'
WEEK_DAYS = 'Weekdays'
WEEK_ENDS = 'Weekends'
ALL_DAYS = 'Alldays'

DAYS_A_MONTH = {'monday': [5, 4, 4, 5, 4, 4, 5, 4, 4, 5, 4, 5],
                'tuesday': [5, 4, 4, 4, 5, 4, 5, 4, 4, 5, 4, 4],
                'wednesday': [5, 4, 4, 4, 5, 4, 4, 5, 4, 5, 4, 4],
                'thursday': [4, 4, 5, 4, 5, 4, 4, 5, 4, 4, 5, 4],
                'friday': [4, 4, 5, 4, 4, 5, 4, 5, 4, 4, 5, 4],
                'saturday': [4, 4, 5, 4, 4, 5, 4, 4, 5, 4, 4, 5],
                'sunday': [4, 4, 4, 5, 4, 4, 5, 4, 5, 4, 4, 5],
                'holiday': [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]}

DAYS_A_YEAR = {'monday': 51,
               'tuesday': 50,
               'wednesday': 50,
               'thursday': 50,
               'friday': 50,
               'saturday': 52,
               'sunday': 52,
               'holiday': 10}

# data types
ANY_NUMBER = 'any_number'
FRACTION = 'fraction'
ON_OFF = 'on_off'
TEMPERATURE = 'temperature'
HUMIDITY = 'humidity'
CONTROL_TYPE = 'control_type'
CONTINUOUS = 'continuous'
DISCRETE = 'discrete'
CONSTANT = 'constant'
INTERNAL_GAINS = 'internal_gains'

# surface types
WALL = 'Wall'
GROUND_WALL = 'Ground wall'
GROUND = 'Ground'
ATTIC_FLOOR = 'Attic floor'
ROOF = 'Roof'
INTERIOR_SLAB = 'Interior slab'
INTERIOR_WALL = 'Interior wall'
VIRTUAL_INTERNAL = 'Virtual internal'
WINDOW = 'Window'
DOOR = 'Door'
SKYLIGHT = 'Skylight'

# functions and usages
RESIDENTIAL = 'residential'
SINGLE_FAMILY_HOUSE = 'single family house'
MULTI_FAMILY_HOUSE = 'multifamily house'
ROW_HOUSE = 'row house'
MID_RISE_APARTMENT = 'mid rise apartment'
HIGH_RISE_APARTMENT = 'high rise apartment'
OFFICE_AND_ADMINISTRATION = 'office and administration'
SMALL_OFFICE = 'small office'
MEDIUM_OFFICE = 'medium office'
LARGE_OFFICE = 'large office'
COURTHOUSE = 'courthouse'
FIRE_STATION = 'fire station'
PENITENTIARY = 'penitentiary'
POLICE_STATION = 'police station'
POST_OFFICE = 'post office'
LIBRARY = 'library'
EDUCATION = 'education'
PRIMARY_SCHOOL = 'primary school'
PRIMARY_SCHOOL_WITH_SHOWER = 'school with shower'
SECONDARY_SCHOOL = 'secondary school'
UNIVERSITY = 'university'
LABORATORY_AND_RESEARCH_CENTER = 'laboratory and research centers'
STAND_ALONE_RETAIL = 'stand alone retail'
HOSPITAL = 'hospital'
OUT_PATIENT_HEALTH_CARE = 'out-patient health care'
HEALTH_CARE = 'health care'
RETIREMENT_HOME_OR_ORPHANAGE = 'retirement home or orphanage'
COMMERCIAL = 'commercial'
STRIP_MALL = 'strip mall'
SUPERMARKET = 'supermarket'
RETAIL_SHOP_WITHOUT_REFRIGERATED_FOOD = 'retail shop without refrigerated food'
RETAIL_SHOP_WITH_REFRIGERATED_FOOD = 'retail shop with refrigerated food'
RESTAURANT = 'restaurant'
QUICK_SERVICE_RESTAURANT = 'quick service restaurant'
FULL_SERVICE_RESTAURANT = 'full service restaurant'
HOTEL = 'hotel'
HOTEL_MEDIUM_CLASS = 'hotel medium class'
SMALL_HOTEL = 'small hotel'
LARGE_HOTEL = 'large hotel'
DORMITORY = 'dormitory'
EVENT_LOCATION = 'event location'
CONVENTION_CENTER = 'convention center'
HALL = 'hall'
GREEN_HOUSE = 'green house'
INDUSTRY = 'industry'
WORKSHOP = 'workshop'
WAREHOUSE = 'warehouse'
WAREHOUSE_REFRIGERATED = 'warehouse refrigerated'
SPORTS_LOCATION = 'sports location'
SPORTS_ARENA = 'sports arena'
GYMNASIUM = 'gymnasium'
MOTION_PICTURE_THEATRE = 'motion picture theatre'
MUSEUM = 'museum'
PERFORMING_ARTS_THEATRE = 'performing arts theatre'
TRANSPORTATION = 'transportation'
AUTOMOTIVE_FACILITY = 'automotive facility'
PARKING_GARAGE = 'parking garage'
RELIGIOUS = 'religious'
NON_HEATED = 'non-heated'
DATACENTER = 'datacenter'

LIGHTING = 'Lights'
OCCUPANCY = 'Occupancy'
APPLIANCES = 'Appliances'
HVAC_AVAILABILITY = 'HVAC Avail'
INFILTRATION = 'Infiltration'
VENTILATION = 'Ventilation'
COOLING_SET_POINT = 'ClgSetPt'
HEATING_SET_POINT = 'HtgSetPt'
EQUIPMENT = 'Equipment'
ACTIVITY = 'Activity'
PEOPLE_ACTIVITY_LEVEL = 'People Activity Level'
DOMESTIC_HOT_WATER = 'Domestic Hot Water'
HEATING = 'Heating'
COOLING = 'Cooling'
ELECTRICITY = 'Electricity'
RENEWABLE = 'Renewable'
WOOD = 'Wood'
GAS = 'Gas'
DIESEL = 'Diesel'
COAL = 'Coal'
AIR = 'Air'
WATER = 'Water'
GEOTHERMAL = 'Geothermal'
DISTRICT_HEATING_NETWORK = 'District Heating'
GRID = 'Grid'
ONSITE_ELECTRICITY = 'Onsite Electricity'
PHOTOVOLTAIC = 'Photovoltaic'
BOILER = 'Boiler'
HEAT_PUMP = 'Heat Pump'
BASEBOARD = 'Baseboard'
CHILLER = 'Chiller'

# Geometry
EPSILON = 0.0000001

# HVAC types
ONLY_HEATING = 'Heating'
ONLY_COOLING = 'Colling'
ONLY_VENTILATION = 'Ventilation'
HEATING_AND_VENTILATION = 'Heating and ventilation'
COOLING_AND_VENTILATION = 'Cooling and ventilation'
HEATING_AND_COLLING = 'Heating and cooling'
FULL_HVAC = 'Heating and cooling and ventilation'

# Floats
MAX_FLOAT = float('inf')
MIN_FLOAT = float('-inf')

# Tools
SRA = 'sra'
INSEL_MEB = 'insel meb'

# Costs units
CURRENCY_PER_SQM = 'currency/m2'
CURRENCY_PER_CBM = 'currency/m3'
CURRENCY_PER_KW = 'currency/kW'
CURRENCY_PER_KWH = 'currency/kWh'
CURRENCY_PER_MONTH = 'currency/month'
CURRENCY_PER_LITRE = 'currency/l'
CURRENCY_PER_KG = 'currency/kg'
CURRENCY_PER_CBM_PER_HOUR = 'currency/(m3/h)'
PERCENTAGE = '%'

# Costs chapters
SUPERSTRUCTURE = 'B_shell'
ENVELOPE = 'D_services'
ALLOWANCES_OVERHEAD_PROFIT = 'Z_allowances_overhead_profit'



"""
Cost fuel
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import Union


class Fuel:
  """
  Fuel class
  """
  def __init__(self, fuel_type,
               fixed_monthly=None,
               fixed_power=None,
               variable=None,
               variable_units=None):

    self._fuel_type = fuel_type
    self._fixed_monthly = fixed_monthly
    self._fixed_power = fixed_power
    self._variable = variable
    self._variable_units = variable_units

  @property
  def type(self):
    """
    Get fuel type
    :return: str
    """
    return self._fuel_type

  @property
  def fixed_monthly(self) -> Union[None, float]:
    """
    Get fixed operational costs in currency per month
    :return: None or float
    """
    return self._fixed_monthly

  @property
  def fixed_power(self) -> Union[None, float]:
    """
    Get fixed operational costs depending on the peak power consumed in currency per month per W
    :return: None or float
    """
    return self._fixed_power/1000

  @property
  def variable(self) -> Union[tuple[None, None], tuple[float, str]]:
    """
    Get variable costs in given units
    :return: None, None or float, str
    """
    return self._variable, self._variable_units

  def to_dictionary(self):
    """Class content to dictionary"""
    content = {'Fuel': {'fuel type': self.type,
                        'fixed operational costs [currency/month]': self.fixed_monthly,
                        'fixed operational costs depending on the peak power consumed [currency/month W]': self.fixed_power,
                        'variable operational costs': self.variable[0],
                        'units': self.variable[1]
                        }
               }

    return content



"""
Geometry helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import math
import sys

import numpy as np
from numpy import ndarray

class GeometryHelper:
  """
  Geometry helper
  """
  @staticmethod
  def to_points_matrix(points):
    """
    Transform a point vector into a point matrix
    :param points: [x, y, z, x, y, z ...]
    :return: [[x,y,z],[x,y,z]...]
    """
    rows = points.size // 3
    points = points.reshape(rows, 3)
    return points

  @staticmethod
  def gml_surface_to_hub(surface):
    """
    Transform citygml surface names into hub names
    """
    if surface == 'WallSurface':
      return 'Wall'
    if surface == 'GroundSurface':
      return 'Ground'
    return 'Roof'

  @staticmethod
  def points_from_string(coordinates) -> np.ndarray:
    """
    Creates a ndarray from a string
    :return: [Point]
    """
    points = np.fromstring(coordinates, dtype=float, sep=' ')
    points = GeometryHelper.to_points_matrix(points)
    return points

  @staticmethod
  def remove_last_point_from_string(points) -> [ndarray]:
    """
    Return the point list without the last element
    :return [ndarray]
    """
    array = points.split(' ')
    res = " "
    return res.join(array[0:len(array) - 3])

  @staticmethod
  def invert_points(points) -> [ndarray]:
    """
    Invert the point list
    :return: [ndarray]
    """
    res = []
    for point in points:
      res.insert(0, point)
    return res

  @staticmethod
  def ground_area(points):
    """
    Get ground surface area in square meters
    :return: float
    """
    # New method to calculate area

    if len(points) < 3:
      sys.stderr.write('Warning: the area of a line or point cannot be calculated 1. Area = 0\n')
      return 0
    alpha = 0
    vec_1 = points[1] - points[0]
    for i in range(2, len(points)):
      vec_2 = points[i] - points[0]
      alpha += GeometryHelper.angle_between_vectors(vec_1, vec_2)
    if alpha == 0:
      sys.stderr.write('Warning: the area of a line or point cannot be calculated 2. Area = 0\n')
      return 0
    #
    horizontal_points = points
    area = 0
    for i in range(0, len(horizontal_points) - 1):
      point = horizontal_points[i]
      next_point = horizontal_points[i + 1]
      area += (next_point[1] + point[1]) / 2 * (next_point[0] - point[0])
    next_point = horizontal_points[0]
    point = horizontal_points[len(horizontal_points) - 1]
    area += (next_point[1] + point[1]) / 2 * (next_point[0] - point[0])
    _area = abs(area)
    return _area

  @staticmethod
  def angle_between_vectors(vec_1, vec_2):
    """
    angle between vectors in radians
    :param vec_1: vector
    :param vec_2: vector
    :return: float
    """
    if np.linalg.norm(vec_1) == 0 or np.linalg.norm(vec_2) == 0:
      sys.stderr.write("Warning: impossible to calculate angle between planes' normal. Return 0\n")
      return 0
    cosine = np.dot(vec_1, vec_2) / np.linalg.norm(vec_1) / np.linalg.norm(vec_2)
    if cosine > 1 and cosine - 1 < 1e-5:
      cosine = 1
    elif cosine < -1 and cosine + 1 > -1e-5:
      cosine = -1
    alpha = math.acos(cosine)
    return alpha



"""
Storeys generation helper
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import math
import sys
from typing import List

import numpy as np

from hub.city_model_structure.attributes.point import Point
from hub.city_model_structure.attributes.polygon import Polygon
from hub.city_model_structure.building_demand.storey import Storey
from hub.city_model_structure.building_demand.surface import Surface
from hub.city_model_structure.building_demand.thermal_zone import ThermalZone
from hub.helpers import constants as cte


class StoreysGeneration:
  """
  StoreysGeneration
  """
  def __init__(self, building, internal_zone, divide_in_storeys=False):
    self._building = building
    self._internal_zone = internal_zone
    self._thermal_zones = []
    self._divide_in_storeys = divide_in_storeys
    self._floor_area = 0
    for ground in building.grounds:
      self._floor_area += ground.perimeter_polygon.area

  @property
  def thermal_zones(self) -> List[ThermalZone]:
    """
    Get subsections of building trimesh by storey in case of no interiors defined
    :return: [Storey]
    """
    number_of_storeys, height = self._calculate_number_storeys_and_height(self._building.average_storey_height,
                                                                          self._building.eave_height,
                                                                          self._building.storeys_above_ground)

    if not self._divide_in_storeys or number_of_storeys == 1:
      storey = Storey('storey_0', self._building.surfaces, [None, None], self._internal_zone.volume,
                      self._internal_zone, self._floor_area)
      for thermal_boundary in storey.thermal_boundaries:
        if thermal_boundary.type != cte.INTERIOR_WALL or thermal_boundary.type != cte.INTERIOR_SLAB:
          # external thermal boundary -> only one thermal zone
          thermal_zones = [storey.thermal_zone]
        else:
          # internal thermal boundary -> two thermal zones
          grad = np.rad2deg(thermal_boundary.parent_surface.inclination)
          if grad >= 170:
            thermal_zones = [storey.thermal_zone, storey.neighbours[0]]
          else:
            thermal_zones = [storey.neighbours[1], storey.thermal_zone]
        thermal_boundary.thermal_zones = thermal_zones
      return [storey.thermal_zone]

    if number_of_storeys == 0:
      raise ArithmeticError('Number of storeys cannot be 0')

    storeys = []
    surfaces_child_last_storey = []
    rest_surfaces = []

    total_volume = 0
    for i in range(0, number_of_storeys - 1):
      name = 'storey_' + str(i)
      surfaces_child = []
      if i == 0:
        neighbours = [None, 'storey_1']
        for surface in self._building.surfaces:
          if surface.type == cte.GROUND:
            surfaces_child.append(surface)
          else:
            rest_surfaces.append(surface)
      else:
        neighbours = ['storey_' + str(i - 1), 'storey_' + str(i + 1)]
      height_division = self._building.lower_corner[2] + height * (i + 1)
      intersections = []
      for surface in rest_surfaces:
        if surface.type == cte.ROOF:
          if height_division >= surface.upper_corner[2] > height_division - height:
            surfaces_child.append(surface)
          else:
            surfaces_child_last_storey.append(surface)
        else:
          surface_child, rest_surface, intersection = surface.divide(height_division)
          surfaces_child.append(surface_child)
          intersections.extend(intersection)
          if i == number_of_storeys - 2:
            surfaces_child_last_storey.append(rest_surface)
      points = []
      for intersection in intersections:
        points.append(intersection[1])
      coordinates = self._intersections_to_coordinates(intersections)
      polygon = Polygon(coordinates)
      ceiling = Surface(polygon, polygon, surface_type=cte.INTERIOR_SLAB)
      surfaces_child.append(ceiling)
      volume = ceiling.perimeter_area * height
      total_volume += volume
      storeys.append(Storey(name, surfaces_child, neighbours, volume, self._internal_zone, self._floor_area))
    name = 'storey_' + str(number_of_storeys - 1)
    neighbours = ['storey_' + str(number_of_storeys - 2), None]
    volume = self._building.volume - total_volume
    if volume < 0:
      raise ArithmeticError('Error in storeys creation, volume of last storey cannot be lower that 0')
    storeys.append(Storey(name, surfaces_child_last_storey, neighbours, volume, self._internal_zone, self._floor_area))

    for storey in storeys:
      for thermal_boundary in storey.thermal_boundaries:
        if thermal_boundary.type != cte.INTERIOR_WALL or thermal_boundary.type != cte.INTERIOR_SLAB:
          # external thermal boundary -> only one thermal zone
          thermal_zones = [storey.thermal_zone]
        else:
          # internal thermal boundary -> two thermal zones
          grad = np.rad2deg(thermal_boundary.parent_surface.inclination)
          if grad >= 170:
            thermal_zones = [storey.thermal_zone, storey.neighbours[0]]
          else:
            thermal_zones = [storey.neighbours[1], storey.thermal_zone]
        thermal_boundary.thermal_zones = thermal_zones

    for storey in storeys:
      self._thermal_zones.append(storey.thermal_zone)

    return self._thermal_zones

  @staticmethod
  def _calculate_number_storeys_and_height(average_storey_height, eave_height, storeys_above_ground):
    if average_storey_height is None:
      if storeys_above_ground is None or storeys_above_ground <= 0:
        sys.stderr.write('Warning: not enough information to divide building into storeys, '
                         'either number of storeys or average storey height must be provided.\n')
        return 0, 0
      number_of_storeys = int(storeys_above_ground)
      height = eave_height / number_of_storeys
    else:
      height = float(average_storey_height)
      if storeys_above_ground is not None:
        number_of_storeys = int(storeys_above_ground)
      else:
        number_of_storeys = math.floor(float(eave_height) / height) + 1
        last_storey_height = float(eave_height) - height*(number_of_storeys-1)
        if last_storey_height < 0.3*height:
          number_of_storeys -= 1
    return number_of_storeys, height

  @staticmethod
  def _intersections_to_coordinates(edges_list):
    # todo: this method is not robust, the while loop needs to be improved
    points = [Point(edges_list[0][0]), Point(edges_list[0][1])]
    found_edges = []
    j = 0
    while j < len(points)-1:
      for i in range(1, len(edges_list)):
        if i not in found_edges:
          point_2 = points[len(points) - 1]
          point_1 = Point(edges_list[i][0])
          found = False
          if point_1.distance_to_point(point_2) <= 1e-10:
            points.append(Point(edges_list[i][1]))
            found_edges.append(i)
            found = True
          if not found:
            point_1 = Point(edges_list[i][1])
            if point_1.distance_to_point(point_2) <= 1e-10:
              points.append(Point(edges_list[i][0]))
              found_edges.append(i)
      j += 1

    points.remove(points[len(points)-1])
    array_points = []
    for point in points:
      array_points.append(point.coordinates)
    return np.array(array_points)



"""
Energy control system module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""


class ControlSystem:
  """
  ControlSystem class
  """
  def __init__(self):
    self._control_type = None

  @property
  def type(self):
    """
    Get control type
    :return: string
    """
    return self._control_type

  @type.setter
  def type(self, value):
    """
    Set control type
    :param value: string
    """
    self._control_type = value



"""
ExportsFactory export a city into several formats
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
"""

import uuid
import datetime
from pathlib import Path
import xmltodict
import hub.helpers.constants as cte


class EnergyAde:
  """
  Export the city to citygml + energy ade
  """
  def __init__(self, city, path):
    self._city = city
    self._path = path
    self._surface_members = None
    self._export()

  def _export(self):
    energy_ade = {
      'core:CityModel': {
        '@xmlns:brid': 'http://www.opengis.net/citygml/bridge/2.0',
        '@xmlns:tran': 'http://www.opengis.net/citygml/transportation/2.0',
        '@xmlns:frn': 'http://www.opengis.net/citygml/cityfurniture/2.0',
        '@xmlns:wtr': 'http://www.opengis.net/citygml/waterbody/2.0',
        '@xmlns:sch': 'http://www.ascc.net/xml/schematron',
        '@xmlns:veg': 'http://www.opengis.net/citygml/vegetation/2.0',
        '@xmlns:xlink': 'http://www.w3.org/1999/xlink',
        '@xmlns:tun': 'http://www.opengis.net/citygml/tunnel/2.0',
        '@xmlns:tex': 'http://www.opengis.net/citygml/texturedsurface/2.0',
        '@xmlns:gml': 'http://www.opengis.net/gml',
        '@xmlns:genobj': 'http://www.opengis.net/citygml/generics/2.0',
        '@xmlns:dem': 'http://www.opengis.net/citygml/relief/2.0',
        '@xmlns:app': 'http://www.opengis.net/citygml/appearance/2.0',
        '@xmlns:luse': 'http://www.opengis.net/citygml/landuse/2.0',
        '@xmlns:xAL': 'urn:oasis:names:tc:ciq:xsdschema:xAL:2.0',
        '@xmlns:xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        '@xmlns:smil20lang': 'http://www.w3.org/2001/SMIL20/Language',
        '@xmlns:pbase': 'http://www.opengis.net/citygml/profiles/base/2.0',
        '@xmlns:smil20': 'http://www.w3.org/2001/SMIL20/',
        '@xmlns:bldg': 'http://www.opengis.net/citygml/building/2.0',
        '@xmlns:energy': "http://www.sig3d.org/citygml/2.0/energy/1.0",
        '@xmlns:core': 'http://www.opengis.net/citygml/2.0',
        '@xmlns:grp': 'http://www.opengis.net/citygml/cityobjectgroup/2.0',
        'gml:boundedBy': {
          'gml:Envelope': {
            '@srsName': self._city.srs_name,
            '@srsDimension': 3,
            'gml:lowerCorner': ' '.join([str(e) for e in self._city.lower_corner]),
            'gml:upperCorner': ' '.join([str(e) for e in self._city.upper_corner])
          }
        }
      }
    }
    buildings = []
    for building in self._city.buildings:
      building_dic = {
        'bldg:Building': {
          '@gml:id': building.name,
          'gml:description': f'Building {building.name} at {self._city.name}',
          'gml:name': f'{building.name}',
          'core:creationDate': datetime.datetime.now().strftime('%Y-%m-%d')
        }
      }
      building_dic = EnergyAde._measures(building, building_dic)
      building_dic = self._building_geometry(building, building_dic, self._city)
      building_dic['bldg:Building']['energy:volume'] = {
        'energy:VolumeType': {
          'energy:type': 'grossVolume',
          'energy:value': {
            '@uom': 'm3',
            '#text': f'{building.volume}'
          }
        }
      }
      building_dic['bldg:Building']['energy:referencePoint'] = {
        'gml:Point': {
          '@srsName': self._city.srs_name,
          '@gml:id': f'GML_{uuid.uuid4()}',
          'gml:Pos': f'{" ".join(map(str, building.centroid))}'
        }
      }
      building_dic['bldg:Building']['energy:thermalZone'] = self._thermal_zones(building, self._city)
      buildings.append(building_dic)

    energy_ade['core:CityModel']['core:cityObjectMember'] = buildings

    file_name = self._city.name + '_ade.gml'
    file_path = Path(self._path / file_name).resolve()
    with open(file_path, 'w', encoding='utf8') as file:
      file.write(xmltodict.unparse(energy_ade, pretty=True, short_empty_elements=True))

  @staticmethod
  def _measures(building, building_dic):
    # todo: this method is only for year and insel need to be generalized
    measures = []
    measure = EnergyAde._measure(building.heating, cte.YEAR, 'Energy demand heating', 'INSEL')
    if measure is not None:
      measures.append(measure)
    measure = EnergyAde._measure(building.cooling, cte.YEAR, 'Energy demand cooling', 'INSEL')
    if measure is not None:
      measures.append(measure)
    if len(measures) != 0:
      building_dic['bldg:Building']['genobj:measureAttribute'] = measures

    demands = []
    for key in building.heating:
      if key != cte.YEAR:
        demand = EnergyAde._demand(building.heating, key, 'Heating energy', 'INSEL')
        demands.append(demand)

    for key in building.cooling:
      if key != cte.YEAR:
        demand = EnergyAde._demand(building.cooling, key, 'Cooling energy', 'INSEL')
        demands.append(demand)
    if len(demands) != 0:
      building_dic['bldg:Building']['energy:demands'] = demands

    return building_dic

  @staticmethod
  def _measure(measure_dict, key_value, name, source):
    measure = None
    if key_value in measure_dict:
      measure = {
        '@name': name,
        'genobj:value': {
          '@uom': 'kWh',
          '#text': ' '.join([str(e / 1000) for e in measure_dict[key_value][source]])
        }
      }
    return measure

  @staticmethod
  def _demand(measure_dict, key_value, description, source):
    demand = {
      'energy:EnergyDemand': {
        '@gml:id': f'GML_{uuid.uuid4()}',
        'energy:energyAmount': {
          'energy:RegularTimeSeries': {
            'energy:variableProperties': {
              'energy:TimeValuesProperties': {
                'energy:acquisitionMethod': 'simulation',
                'energy:source': source,
                'energy:thematicDescription': description,
              },
              'energy:timeInterval': {
                '@unit': key_value,
                '#text': '1',
              },
              'energy:values': {
                '@uom': 'kWh',
                '#text': ' '.join([str(float(e) / 1000) for e in measure_dict[key_value][source]])
              }
            }
          }
        },
        'energy:endUse': 'spaceHeating'
      }
    }
    return demand

  def _building_geometry(self, building, building_dic, city):

    building_dic['bldg:Building']['bldg:function'] = building.function
    building_dic['bldg:Building']['bldg:usage'] = building.usages_percentage
    building_dic['bldg:Building']['bldg:yearOfConstruction'] = building.year_of_construction
    building_dic['bldg:Building']['bldg:roofType'] = building.roof_type
    building_dic['bldg:Building']['bldg:measuredHeight'] = {
      '@uom': 'm',
      '#text': f'{building.max_height}'
    }
    building_dic['bldg:Building']['bldg:storeysAboveGround'] = building.storeys_above_ground

    if city.level_of_detail.geometry == 1:
      building_dic = self._lod1(building, building_dic, city)
    elif city.level_of_detail.geometry == 2:
      building_dic = self._lod2(building, building_dic, city)
    else:
      raise NotImplementedError('Only lod 1 and 2 can be exported')
    return building_dic

  def _lod1(self, building, building_dic, city):
    self._surface_members = []
    boundaries = [{
      'gml:Envelope': {
        '@srsName': city.srs_name,
        '@srsDimension': 3,
        'gml:lowerCorner': ' '.join([str(e) for e in city.lower_corner]),
        'gml:upperCorner': ' '.join([str(e) for e in city.upper_corner])
      }}]
    for surface in building.surfaces:
      surface_member = {'@xlink:href': f'#PolyId{surface.name}'}
      self._surface_members.append(surface_member)
      if surface.type == 'Wall':
        surface_type = 'bldg:WallSurface'
      elif surface.type == 'Ground':
        surface_type = 'bldg:GroundSurface'
      else:
        surface_type = 'bldg:RoofSurface'
      surface_dic = {
        surface_type: {
          '@gml:id': f'GML_{uuid.uuid4()}',
          'gml:name': f'{surface.name} ({surface.type})',
          'gml:boundedBy': {
            'gml:Envelope': {
              '@srsName': city.srs_name,
              'gml:lowerCorner': f'{surface.lower_corner[0]} {surface.lower_corner[1]}'
                                 f' {surface.lower_corner[2]}',
              'gml:upperCorner': f'{surface.upper_corner[0]} {surface.upper_corner[1]}'
                                 f' {surface.upper_corner[2]}'
            }
          },
          'bldg:lod1MultiSurface': {
            'gml:MultiSurface': {
              '@srsName': city.srs_name,
              '@gml:id': f'GML_{uuid.uuid4()}',
              'surfaceMember': {
                'gml:Polygon': {
                  '@srsName': city.srs_name,
                  '@gml:id': f'PolyId{surface.name}',
                  'gml:exterior': {
                    'gml:LinearRing': {
                      '@gml:id': f'PolyId{surface.name}_0',
                      'gml:posList': {
                        '@srsDimension': '3',
                        '@count': len(surface.solid_polygon.coordinates) + 1,
                        '#text': f'{" ".join(map(str, surface.solid_polygon.points_list))} '
                                 f'{" ".join(map(str, surface.solid_polygon.coordinates[0]))}'
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
      boundaries.append(surface_dic)
    building_dic['bldg:Building']['bldg:lod1Solid'] = {
      'gml:Solid': {
        '@gml:id': f'GML_{uuid.uuid4()}',
        'gml:exterior': {
          'gml:CompositeSurface': {
            '@srsName': city.srs_name,
            '@gml:id': f'GML_{uuid.uuid4()}',
            'gml:surfaceMember': self._surface_members
          }
        }
      }
    }

    building_dic['bldg:Building']['gml:boundedBy'] = boundaries
    return building_dic

  def _lod2(self, building, building_dic, city):
    self._surface_members = []
    boundaries = [{
      'gml:Envelope': {
        '@srsName': city.srs_name,
        '@srsDimension': 3,
        'gml:lowerCorner': ' '.join([str(e) for e in city.lower_corner]),
        'gml:upperCorner': ' '.join([str(e) for e in city.upper_corner])
      }}]
    for surface in building.surfaces:
      surface_member = {'@xlink:href': f'#PolyId{surface.name}'}
      self._surface_members.append(surface_member)
      if surface.type == 'Wall':
        surface_type = 'bldg:WallSurface'
      elif surface.type == 'Ground':
        surface_type = 'bldg:GroundSurface'
      else:
        surface_type = 'bldg:RoofSurface'
      surface_dic = {
        surface_type: {
          '@gml:id': f'GML_{uuid.uuid4()}',
          'gml:name': f'{surface.name} ({surface.type})',
          'gml:boundedBy': {
            'gml:Envelope': {
              '@srsName': city.srs_name,
              'gml:lowerCorner': f'{surface.lower_corner[0]} {surface.lower_corner[1]}'
                                 f' {surface.lower_corner[2]}',
              'gml:upperCorner': f'{surface.upper_corner[0]} {surface.upper_corner[1]}'
                                 f' {surface.upper_corner[2]}'
            }
          },
          'bldg:lod2MultiSurface': {
            'gml:MultiSurface': {
              '@srsName': city.srs_name,
              '@gml:id': f'GML_{uuid.uuid4()}',
              'surfaceMember': {
                'gml:Polygon': {
                  '@srsName': city.srs_name,
                  '@gml:id': f'PolyId{surface.name}',
                  'gml:exterior': {
                    'gml:LinearRing': {
                      '@gml:id': f'PolyId{surface.name}_0',
                      'gml:posList': {
                        '@srsDimension': '3',
                        '@count': len(surface.solid_polygon.coordinates) + 1,
                        '#text': f'{" ".join(map(str, surface.solid_polygon.points_list))} '
                                 f'{" ".join(map(str, surface.solid_polygon.coordinates[0]))}'
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
      boundaries.append(surface_dic)
    building_dic['bldg:Building']['bldg:lod2Solid'] = {
      'gml:Solid': {
        '@gml:id': f'GML_{uuid.uuid4()}',
        'gml:exterior': {
          'gml:CompositeSurface': {
            '@srsName': city.srs_name,
            '@gml:id': f'GML_{uuid.uuid4()}',
            'gml:surfaceMember': self._surface_members
          }
        }
      }
    }

    building_dic['bldg:Building']['gml:boundedBy'] = boundaries
    return building_dic

  def _thermal_zones(self, building, city):
    thermal_zones = []
    for internal_zone in building.internal_zones:
      for index, thermal_zone in enumerate(internal_zone.thermal_zones):
        usages = []
        for usage in internal_zone.usages:
          usages.append({'@xlink:href': f'#GML_{usage.id}'})
        thermal_zone_dic = {
          'energy:ThermalZone': {
            '@gml:id': f'GML_{thermal_zone.id}',
            'gml:name': f'Thermal zone {index} in {building.name} building',
            'energy:contains': [],
            'energy:floorArea': {
              'energy:FloorArea': {
                'energy:type': 'grossFloorArea',
                'energy:value': {
                  '@uom': 'm2',
                  '#text': f'{thermal_zone.footprint_area}'
                }
              }
            },
            'energy:volume': {
              'energy:VolumeType': {
                'energy:type': 'grossVolume',
                'energy:value': {
                  '@uom': 'm3',
                  '#text': f'{thermal_zone.volume}'
                }
              }
            },
            'energy:isCooled': f'{building.is_conditioned}'.lower(),
            'energy:isHeated': f'{building.is_conditioned}'.lower(),
            'energy:volumeGeometry': {
              'gml:Solid': {
                '@gml:id': f'GML_{uuid.uuid4()}',
                'gml:exterior': {
                  'gml:CompositeSurface': {
                    '@srsName': f'{city.srs_name}',
                    '@gml:id': f'GML_{uuid.uuid4()}',
                    'gml:surfaceMember': self._surface_members
                  }
                }
              }
            },
            'energy:boundedBy': self._thermal_boundaries(city, thermal_zone)
          }
        }
        thermal_zone_dic['energy:ThermalZone']['energy:contains'] = usages
        thermal_zones.append(thermal_zone_dic)
    return thermal_zones

  @staticmethod
  def _thermal_boundaries(city, thermal_zone):
    thermal_boundaries = []
    for thermal_boundary in thermal_zone.thermal_boundaries:
      thermal_boundary_dic = {
        '@gml:id': f'GML_{uuid.uuid4()}',
        'gml:name': f'{thermal_boundary.construction_name}',
        'energy:thermalBoundaryType': thermal_boundary.type,
        'energy:azimuth': {
          '@uom': 'rad',
          '#text': f'{thermal_boundary.parent_surface.azimuth}'
        },
        'energy:inclination': {
          '@uom': 'rad',
          '#text': f'{thermal_boundary.parent_surface.inclination}'
        },
        'energy:area': {
          '@uom': 'm2',
          '#text': f'{thermal_boundary.opaque_area}'
        },
        'energy:surfaceGeometry': {
          'gml:MultiSurface': {
            '@srsName': city.srs_name,
            '@gml:id': f'GML_{uuid.uuid4()}',
            'gml:surfaceMember': {
              'gml:Polygon': {
                '@srsName': city.srs_name,
                '@gml:id': f'GML_{uuid.uuid4()}',
                'gml:exterior': {
                  'gml:LinearRing': {
                    '@gml:id': f'GML_{uuid.uuid4()}',
                    'gml:posList': {
                      '@srsDimension': '3',
                      '@count': len(thermal_boundary.parent_surface.solid_polygon.coordinates) + 1,
                      '#text': f'{" ".join(map(str, thermal_boundary.parent_surface.solid_polygon.points_list))} '
                               f'{" ".join(map(str, thermal_boundary.parent_surface.solid_polygon.coordinates[0]))}'
                    }
                  }
                }
              }
            }
          }
        }
      }
      construction = []
      opening_construction = []
      if thermal_boundary.layers is not None:
        construction.append(uuid.uuid4())
        thermal_boundary_dic['energy:construction'] = {
          '@xlink:href': f'#GML_{construction[0]}'
        }
      if thermal_boundary.thermal_openings is not None:
        for _ in thermal_boundary.thermal_openings:
          opening_construction.append(uuid.uuid4())
          thermal_boundary_dic['energy:contains'] = {
            'energy:ThermalOpening': {
              '@gml:id': f'GML_{uuid.uuid4()}',
              'energy:construction': f'#GML_{opening_construction}',
              'energy:surfaceGeometry': {

              }
            }
          }
      thermal_boundaries.append(thermal_boundary_dic)
    return thermal_boundaries



"""
TestInselExports test
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""
import logging
from pathlib import Path
from unittest import TestCase
import pandas as pd
import hub.helpers.constants as cte
from hub.helpers.monthly_values import MonthlyValues
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.construction_factory import ConstructionFactory
from hub.imports.usage_factory import UsageFactory
from hub.exports.energy_building_exports_factory import EnergyBuildingsExportsFactory
from hub.imports.weather_factory import WeatherFactory


class TestExports(TestCase):
  """
  TestExports class contains the unittest for export functionality
  """
  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    self._city = None
    self._complete_city = None
    self._example_path = (Path(__file__).parent / 'tests_data').resolve()
    self._output_path = (Path(__file__).parent / 'tests_outputs').resolve()

  def _get_citygml(self, file):
    file_path = (self._example_path / file).resolve()
    self._city = GeometryFactory('citygml', path=file_path).city
    self.assertIsNotNone(self._city, 'city is none')
    return self._city

  @property
  def _read_sra_file(self) -> []:
    path = (self._example_path / "one_building_in_kelowna_sra_SW.out").resolve()
    _results = pd.read_csv(path, sep='\s+', header=0)
    id_building = ''
    header_building = []
    _radiation = []
    for column in _results.columns.values:
      if id_building != column.split(':')[1]:
        id_building = column.split(':')[1]
        if len(header_building) > 0:
          _radiation.append(pd.concat([MonthlyValues().month_hour, _results[header_building]], axis=1))
        header_building = [column]
      else:
        header_building.append(column)
    _radiation.append(pd.concat([MonthlyValues().month_hour, _results[header_building]], axis=1))
    return _radiation

  def _set_irradiance_surfaces(self, city):
    """
    saves in building surfaces the correspondent irradiance at different time-scales depending on the mode
    if building is None, it saves all buildings' surfaces in file, if building is specified, it saves only that
    specific building values
    :parameter city: city
    :return: none
    """
    city.level_of_detail.surface_radiation = 2
    for radiation in self._read_sra_file:
      city_object_name = radiation.columns.values.tolist()[1].split(':')[1]
      building = city.city_object(city_object_name)
      for column in radiation.columns.values:
        if column == cte.MONTH:
          continue
        header_id = column
        surface_id = header_id.split(':')[2]
        surface = building.surface_by_id(surface_id)
        new_value = pd.DataFrame(radiation[[header_id]].to_numpy(), columns=['sra'])
        surface.global_irradiance[cte.HOUR] = new_value
        month_new_value = MonthlyValues().get_mean_values(new_value)
        surface.global_irradiance[cte.MONTH] = month_new_value

  def test_insel_monthly_energy_balance_export(self):
    """
    export to Insel MonthlyEnergyBalance
    """
    city = self._get_citygml('one_building_in_kelowna.gml')
    WeatherFactory('epw', city).enrich()
    for building in city.buildings:
      building.external_temperature[cte.MONTH] = MonthlyValues().\
        get_mean_values(building.external_temperature[cte.HOUR][['epw']])
    self._set_irradiance_surfaces(city)

    for building in city.buildings:
      self.assertIsNotNone(building.external_temperature[cte.MONTH], f'building {building.name} '
                                                                     f'external_temperature is none')
      for surface in building.surfaces:
        if surface.type != 'Ground':
          self.assertIsNotNone(surface.global_irradiance[cte.MONTH], f'surface in building {building.name} '
                                                                     f'global_irradiance is none')

    for building in city.buildings:
      building.year_of_construction = 2006
      if building.function is None:
        building.function = 'large office'
      building.attic_heated = 0
      building.basement_heated = 0
    ConstructionFactory('nrel', city).enrich()
    UsageFactory('comnet', city).enrich()

    # parameters written:
    for building in city.buildings:
      self.assertIsNotNone(building.volume, f'building {building.name} volume is none')
      self.assertIsNotNone(building.average_storey_height, f'building {building.name} average_storey_height is none')
      self.assertIsNotNone(building.storeys_above_ground, f'building {building.name} storeys_above_ground is none')
      self.assertIsNotNone(building.attic_heated, f'building {building.name} attic_heated is none')
      self.assertIsNotNone(building.basement_heated, f'building {building.name} basement_heated is none')
      for internal_zone in building.internal_zones:
        self.assertIsNotNone(internal_zone.area, f'internal zone {internal_zone.id} area is none')
        for thermal_zone in internal_zone.thermal_zones:
          self.assertIsNotNone(thermal_zone.indirectly_heated_area_ratio, f'thermal zone {thermal_zone.id} '
                                                                          f'indirectly_heated_area_ratio is none')
          self.assertIsNotNone(thermal_zone.effective_thermal_capacity, f'thermal zone {thermal_zone.id} '
                                                                        f'effective_thermal_capacity is none')
          self.assertIsNotNone(thermal_zone.additional_thermal_bridge_u_value, f'thermal zone {thermal_zone.id} '
                                                                               f'additional_thermal_bridge_u_value '
                                                                               f'is none')
          self.assertIsNotNone(thermal_zone.total_floor_area, f'thermal zone {thermal_zone.id} '
                                                              f'total_floor_area is none')
          for thermal_boundary in thermal_zone.thermal_boundaries:
            self.assertIsNotNone(thermal_boundary.type)
            self.assertIsNotNone(thermal_boundary.opaque_area)
            self.assertIsNotNone(thermal_boundary.window_ratio)
            self.assertIsNotNone(thermal_boundary.u_value)
            self.assertIsNotNone(thermal_boundary.thermal_openings)
            if thermal_boundary.type is not cte.GROUND:
              self.assertIsNotNone(thermal_boundary.parent_surface.short_wave_reflectance)

        for usage in internal_zone.usages:
          self.assertIsNotNone(usage.percentage, f'usage zone {usage.name} percentage is none')
          self.assertIsNotNone(usage.internal_gains, f'usage zone {usage.name} internal_gains is none')
          self.assertIsNotNone(usage.thermal_control, f'usage zone {usage.name} thermal_control is none')
          self.assertIsNotNone(usage.hours_day, f'usage zone {usage.name} hours_day is none')
          self.assertIsNotNone(usage.days_year, f'usage zone {usage.name} days_year is none')
          self.assertIsNotNone(
            usage.mechanical_air_change,
            f'usage zone {usage.name} mechanical_air_change is none'
          )
    # export files
    try:
      EnergyBuildingsExportsFactory('insel_monthly_energy_balance', city, self._output_path, 'MEB_Montreal').export()
    except Exception as err:
      logging.exception(err)
      self.fail("Insel MonthlyEnergyBalance ExportsFactory raised ExceptionType unexpectedly!")



"""
Energy emission system definition
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from hub.city_model_structure.energy_systems.generic_emission_system import GenericEmissionSystem


class EmissionSystem:
  """
  EmissionSystem class
  """
  def __init__(self):
    self._generic_emission_system = None

  @property
  def generic_emission_system(self) -> GenericEmissionSystem:
    """
    Get associated generic_emission_system
    :return: GenericEmissionSystem
    """
    return self._generic_emission_system

  @generic_emission_system.setter
  def generic_emission_system(self, value):
    """
    Set associated
    :param value: GenericEmissionSystem
    """
    self._generic_emission_system = value



"""
Montreal custom energy system importer
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import logging
import copy

from pandas import DataFrame

from hub.catalog_factories.energy_systems_catalog_factory import EnergySystemsCatalogFactory
from hub.city_model_structure.energy_systems.generic_distribution_system import GenericDistributionSystem
from hub.city_model_structure.energy_systems.generic_energy_system import GenericEnergySystem
from hub.city_model_structure.energy_systems.generic_generation_system import GenericGenerationSystem
from hub.city_model_structure.energy_systems.energy_system import EnergySystem
from hub.city_model_structure.energy_systems.generation_system import GenerationSystem
from hub.city_model_structure.energy_systems.distribution_system import DistributionSystem
from hub.city_model_structure.energy_systems.emission_system import EmissionSystem
from hub.helpers.dictionaries import Dictionaries


class MontrealCustomEnergySystemParameters:
  """
  MontrealCustomEnergySystemParameters class
  """

  def __init__(self, city):
    self._city = city

  def enrich_buildings(self):
    """
    Returns the city with the system parameters assigned to the buildings
    :return:
    """
    city = self._city
    montreal_custom_catalog = EnergySystemsCatalogFactory('montreal_custom').catalog
    if city.energy_systems_connection_table is None:
      _energy_systems_connection_table = DataFrame(columns=['Energy System Type', 'Building'])
    else:
      _energy_systems_connection_table = city.energy_systems_connection_table
    if city.generic_energy_systems is None:
      _generic_energy_systems = {}
    else:
      _generic_energy_systems = city.generic_energy_systems
    for building in city.buildings:
      archetype_name = f'{building.energy_systems_archetype_name}_lod1.0'
      try:
        archetype = self._search_archetypes(montreal_custom_catalog, archetype_name)
      except KeyError:
        logging.error('Building %s has unknown energy system archetype for system name %s', building.name,
                      archetype_name)
        continue

      _energy_systems_connection_table, _generic_energy_systems \
        = self._create_generic_systems(archetype, building,
                                       _energy_systems_connection_table, _generic_energy_systems)
    city.energy_systems_connection_table = _energy_systems_connection_table
    city.generic_energy_systems = _generic_energy_systems

    self._associate_energy_systems(city)

  @staticmethod
  def _search_archetypes(catalog, name):
    archetypes = catalog.entries('archetypes')
    for building_archetype in archetypes:
      if str(name) == str(building_archetype.name):
        return building_archetype
    raise KeyError('archetype not found')

  @staticmethod
  def _create_generic_systems(archetype, building,
                              _energy_systems_connection_table, _generic_energy_systems):
    building_systems = []
    data = [archetype.name, building.name]
    _energy_systems_connection_table.loc[len(_energy_systems_connection_table)] = data
    for system in archetype.systems:
      energy_system = GenericEnergySystem()
      _hub_demand_types = []
      for demand_type in system.demand_types:
        _hub_demand_types.append(Dictionaries().montreal_demand_type_to_hub_energy_demand_type[demand_type])
      energy_system.name = system.name
      energy_system.demand_types = _hub_demand_types
      _generation_system = GenericGenerationSystem()
      archetype_generation_equipment = system.generation_system
      _type = str(system.name).split('_', maxsplit=1)[0]
      _generation_system.type = Dictionaries().montreal_system_to_hub_energy_generation_system[
        _type]
      _generation_system.fuel_type = archetype_generation_equipment.fuel_type
      _generation_system.source_types = archetype_generation_equipment.source_types
      _generation_system.heat_efficiency = archetype_generation_equipment.heat_efficiency
      _generation_system.cooling_efficiency = archetype_generation_equipment.cooling_efficiency
      _generation_system.electricity_efficiency = archetype_generation_equipment.electricity_efficiency
      _generation_system.source_temperature = archetype_generation_equipment.source_temperature
      _generation_system.source_mass_flow = archetype_generation_equipment.source_mass_flow
      _generation_system.storage = archetype_generation_equipment.storage
      _generation_system.auxiliary_equipment = None

      energy_system.generation_system = _generation_system

      _distribution_system = GenericDistributionSystem()
      archetype_distribution_equipment = system.distribution_system
      _distribution_system.type = archetype_distribution_equipment.type
      _distribution_system.supply_temperature = archetype_distribution_equipment.supply_temperature
      _distribution_system.distribution_consumption_fix_flow = \
        archetype_distribution_equipment.distribution_consumption_fix_flow
      _distribution_system.distribution_consumption_variable_flow = \
        archetype_distribution_equipment.distribution_consumption_variable_flow
      _distribution_system.heat_losses = archetype_distribution_equipment.heat_losses

      energy_system.distribution_system = _distribution_system

      building_systems.append(energy_system)
    if archetype.name not in _generic_energy_systems:
      _generic_energy_systems[archetype.name] = building_systems

    return _energy_systems_connection_table, _generic_energy_systems

  @staticmethod
  def _associate_energy_systems(city):
    energy_systems_connection = city.energy_systems_connection_table
    for building in city.buildings:
      _building_energy_systems = []
      energy_systems = energy_systems_connection['Energy System Type'][
        energy_systems_connection['Building'] == building.name]
      for energy_system in energy_systems:
        if str(energy_system) == 'nan':
          break
        _generic_building_energy_systems = city.generic_energy_systems[energy_system]
        for _generic_building_energy_system in _generic_building_energy_systems:
          _building_energy_equipment = EnergySystem()
          _building_energy_equipment.name = _generic_building_energy_system.name
          _building_energy_equipment.demand_types = _generic_building_energy_system.demand_types

          _building_distribution_system = DistributionSystem()
          _building_distribution_system.generic_distribution_system = \
            copy.deepcopy(_generic_building_energy_system.distribution_system)
          _building_emission_system = EmissionSystem()
          _building_emission_system.generic_emission_system = \
            copy.deepcopy(_generic_building_energy_system.emission_system)
          _building_generation_system = GenerationSystem()
          _building_generation_system.generic_generation_system = \
            copy.deepcopy(_generic_building_energy_system.generation_system)

          _building_energy_equipment.generation_system = _building_generation_system
          _building_energy_equipment.distribution_system = _building_distribution_system
          _building_energy_equipment.emission_system = _building_emission_system

          _building_energy_systems.append(_building_energy_equipment)
      building.energy_systems = _building_energy_systems



"""
Test EnergySystemsFactory and various heatpump models
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Peter Yefi peteryefi@gmail.com
"""
from unittest import TestCase
from hub.imports.geometry_factory import GeometryFactory
from hub.imports.energy_systems_factory import EnergySystemsFactory
from hub.exports.energy_systems_factory import EnergySystemsExportFactory
from hub.imports.results_factory import ResultFactory
import os
from pandas.core.series import Series

# User defined parameters
user_input = {
  'StartYear': 2020,
  'EndYear': 2021,
  'MaximumHPEnergyInput': 8000,
  'HoursOfStorageAtMaxDemand': 1,
  'BuildingSuppTemp': 40,
  'TemperatureDifference': 15,
  'FuelLHV': 47100,
  'FuelPrice': 0.12,
  'FuelEF': 1887,
  'FuelDensity': 0.717,
  'HPSupTemp': 60
}


class TestHeatPumpResults(TestCase):
  """
  TestHeatPumpResults
  """

  def setUp(self) -> None:
    """
    Test setup
    :return: None
    """
    city_file = "tests_data/C40_Final.gml"
    self._output_path = "tests_data/as_user_output.csv"
    self._city = GeometryFactory('citygml', path=city_file).city
    EnergySystemsFactory('air_source_hp', self._city).enrich()

  def test_air_source_series_heat_pump_012_results(self):
    EnergySystemsExportFactory(city=self._city, handler=user_input, hp_model='012',
                               output_path=self._output_path).export()
    ResultFactory('heat_pump', self._city, self._output_path, '012').enrich()

    for energy_system in self._city.energy_systems:
      self.assertIsNone(energy_system.water_to_water_hp)
      if energy_system.air_source_hp.model == '012':
        self.assertIsInstance(energy_system.air_source_hp.hp_monthly_fossil_consumption, Series)
        self.assertEqual(energy_system.air_source_hp.hp_monthly_fossil_consumption.iloc[5], 1.51325583)
        self.assertEqual(energy_system.air_source_hp.hp_monthly_fossil_consumption.iloc[12], 35.853598782915)

  def test_air_source_series_heat_pump_015_results(self):
    EnergySystemsExportFactory(city=self._city, handler=user_input, hp_model='140',
                               output_path=self._output_path).export()
    ResultFactory('heat_pump', self._city, self._output_path, '140').enrich()

    for energy_system in self._city.energy_systems:
      self.assertIsNone(energy_system.water_to_water_hp)
      if energy_system.air_source_hp.model == '140':
        self.assertIsInstance(energy_system.air_source_hp.hp_monthly_fossil_consumption, Series)
        self.assertEqual(energy_system.air_source_hp.hp_monthly_fossil_consumption.iloc[0], 7.91282225)
        self.assertEqual(energy_system.air_source_hp.hp_monthly_fossil_consumption.iloc[2], 0.068873927)

  def tearDown(self) -> None:
    try:
      os.remove(self._output_path)
    except OSError:
      pass



"""
Energy System catalog equipment
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

from typing import Union

from hub.catalog_factories.data_models.energy_systems.generation_system import GenerationSystem
from hub.catalog_factories.data_models.energy_systems.distribution_system import DistributionSystem
from hub.catalog_factories.data_models.energy_systems.emission_system import EmissionSystem


class System:
  """
  System class
  """
  def __init__(self,
               lod,
               system_id,
               name,
               demand_types,
               generation_system,
               distribution_system,
               emission_system):

    self._lod = lod
    self._system_id = system_id
    self._name = name
    self._demand_types = demand_types
    self._generation_system = generation_system
    self._distribution_system = distribution_system
    self._emission_system = emission_system

  @property
  def lod(self):
    """
    Get level of detail of the catalog
    :return: string
    """
    return self._lod

  @property
  def id(self):
    """
    Get equipment id
    :return: string
    """
    return self._system_id

  @property
  def name(self):
    """
    Get name
    :return: string
    """
    return f'{self._name}_lod{self._lod}'

  @property
  def demand_types(self):
    """
    Get demand able to cover from [heating, cooling, domestic_hot_water, electricity]
    :return: [string]
    """
    return self._demand_types

  @property
  def generation_system(self) -> GenerationSystem:
    """
    Get generation system
    :return: GenerationSystem
    """
    return self._generation_system

  @property
  def distribution_system(self) -> Union[None, DistributionSystem]:
    """
    Get distribution system
    :return: DistributionSystem
    """
    return self._distribution_system

  @property
  def emission_system(self) -> Union[None, EmissionSystem]:
    """
    Get emission system
    :return: EmissionSystem
    """
    return self._emission_system

  def to_dictionary(self):
    """Class content to dictionary"""
    _distribution_system = None
    if self.distribution_system is not None:
      _distribution_system = self.distribution_system.to_dictionary()
    _emission_system = None
    if self.emission_system is not None:
      _emission_system = self.emission_system.to_dictionary()
    content = {'Layer': {'id': self.id,
                         'name': self.name,
                         'level of detail': self.lod,
                         'demand types': self.demand_types,
                         'generation system': self.generation_system.to_dictionary(),
                         'distribution system': _distribution_system,
                         'emission system': _emission_system
                         }
               }
    return content



"""
City module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Guille Gutierrez guillermo.gutierrezmorote@concordia.ca
Code contributors: Peter Yefi peteryefi@gmail.com
"""
from __future__ import annotations

import bz2
import copy
import logging
import math
import pickle
import sys
import pathlib
from pathlib import Path
from typing import List, Union

import pyproj
from pandas import DataFrame
from pyproj import Transformer

from hub.city_model_structure.building import Building
from hub.city_model_structure.buildings_cluster import BuildingsCluster
from hub.city_model_structure.city_object import CityObject
from hub.city_model_structure.city_objects_cluster import CityObjectsCluster
from hub.city_model_structure.energy_system import EnergySystem
from hub.city_model_structure.iot.station import Station
from hub.city_model_structure.level_of_detail import LevelOfDetail
from hub.city_model_structure.parts_consisting_building import PartsConsistingBuilding
from hub.helpers.geometry_helper import GeometryHelper
from hub.helpers.location import Location
import hub.helpers.constants as cte


class City:
  """
  City class
  """

  def __init__(self, lower_corner, upper_corner, srs_name):
    self._name = None
    self._lower_corner = lower_corner
    self._upper_corner = upper_corner
    self._buildings = []
    self._srs_name = srs_name
    self._location = None
    self._country_code = None
    self._climate_reference_city = None
    self._climate_file = None
    self._latitude = None
    self._longitude = None
    self._time_zone = None
    self._buildings_clusters = None
    self._parts_consisting_buildings = None
    self._city_objects_clusters = None
    self._city_objects = None
    self._energy_systems = None
    self._fuels = None
    self._stations = []
    self._level_of_detail = LevelOfDetail()
    self._city_objects_dictionary = {}
    self._city_objects_alias_dictionary = {}
    self._energy_systems_connection_table = None
    self._generic_energy_systems = None

  def _get_location(self) -> Location:
    if self._location is None:
      gps = pyproj.CRS('EPSG:4326')  # LatLon with WGS84 datum used by GPS units and Google Earth
      try:
        if self._srs_name in GeometryHelper.srs_transformations:
          self._srs_name = GeometryHelper.srs_transformations[self._srs_name]
        input_reference = pyproj.CRS(self.srs_name)  # Projected coordinate system from input data
      except pyproj.exceptions.CRSError as err:
        logging.error('Invalid projection reference system, please check the input data.')
        raise pyproj.exceptions.CRSError from err
      transformer = Transformer.from_crs(input_reference, gps)
      coordinates = transformer.transform(self.lower_corner[0], self.lower_corner[1])
      self._location = GeometryHelper.get_location(coordinates[0], coordinates[1])
    return self._location

  @property
  def country_code(self):
    """
    Get city country code
    :return: str
    """
    return self._get_location().country

  @property
  def region_code(self):
    """
    Get city region name
    :return: str
    """
    return self._get_location().region_code

  @property
  def location(self) -> Location:
    """
    Get city location
    :return: Location
    """
    return self._get_location().city

  @property
  def name(self):
    """
    Get city name
    :return: str
    """
    if self._name is None:
      return self._get_location().city
    return self._name

  @property
  def climate_reference_city(self) -> Union[None, str]:
    """
    Get the name for the climatic information reference city
    :return: None or str
    """
    if self._climate_reference_city is None:
      self._climate_reference_city = self._get_location().city
    return self._climate_reference_city

  @climate_reference_city.setter
  def climate_reference_city(self, value):
    """
    Set the name for the climatic information reference city
    :param value: str
    """
    self._climate_reference_city = str(value)

  @property
  def climate_file(self) -> Union[None, Path]:
    """
    Get the climate file full path
    :return: None or Path
    """
    return self._climate_file

  @climate_file.setter
  def climate_file(self, value):
    """
    Set the climate file full path
    :param value: Path
    """
    if value is not None:
      self._climate_file = Path(value)

  @property
  def city_objects(self) -> Union[List[CityObject], None]:
    """
    Get the city objects belonging to the city
    :return: None or [CityObject]
    """
    if self._city_objects is None:
      if self.city_objects_clusters is None:
        self._city_objects = []
      else:
        self._city_objects = self.city_objects_clusters
      if self.buildings is not None:
        for building in self.buildings:
          self._city_objects.append(building)
      if self.energy_systems is not None:
        for energy_system in self.energy_systems:
          self._city_objects.append(energy_system)
    return self._city_objects

  @property
  def buildings(self) -> Union[List[Building], None]:
    """
    Get the buildings belonging to the city
    :return: None or [Building]
    """
    return self._buildings

  @property
  def lower_corner(self) -> List[float]:
    """
    Get city lower corner
    :return: [x,y,z]
    """
    return self._lower_corner

  @property
  def upper_corner(self) -> List[float]:
    """
    Get city upper corner
    :return: [x,y,z]
    """
    return self._upper_corner

  def city_object(self, name) -> Union[CityObject, None]:
    """
    Retrieve the city CityObject with the given name
    :param name:str
    :return: None or CityObject
    """
    if name in self._city_objects_dictionary:
      return self.buildings[self._city_objects_dictionary[name]]
    return None

  def building_alias(self, alias) -> list[Building | list[Building]] | None:
    """
    Retrieve the city CityObject with the given alias alias
    :alert: Building alias is not guaranteed to be unique
    :param alias:str
    :return: None or [CityObject]
    """
    if alias in self._city_objects_alias_dictionary:
      return [self.buildings[i] for i in self._city_objects_alias_dictionary[alias]]
    return None

  def add_building_alias(self, building, alias):
    """
    Add an alias to the building
    """
    building_index = self._city_objects_dictionary[building.name]
    if alias in self._city_objects_alias_dictionary:
      self._city_objects_alias_dictionary[alias].append(building_index)
    else:
      self._city_objects_alias_dictionary[alias] = [building_index]

  def add_city_object(self, new_city_object):
    """
    Add a CityObject to the city
    :param new_city_object:CityObject
    :return: None or not implemented error
    """
    if new_city_object.type == 'building':
      if self._buildings is None:
        self._buildings = []
      self._buildings.append(new_city_object)
      self._city_objects_dictionary[new_city_object.name] = len(self._buildings) - 1
      if new_city_object.aliases is not None:
        for alias in new_city_object.aliases:
          if alias in self._city_objects_alias_dictionary:
            self._city_objects_alias_dictionary[alias].append(len(self._buildings) - 1)
          else:
            self._city_objects_alias_dictionary[alias] = [len(self._buildings) - 1]
    elif new_city_object.type == 'energy_system':
      if self._energy_systems is None:
        self._energy_systems = []
      self._energy_systems.append(new_city_object)
    else:
      raise NotImplementedError(new_city_object.type)

  def remove_city_object(self, city_object):
    """
    Remove a CityObject from the city
    :param city_object:CityObject
    :return: None
    """
    if city_object.type != 'building':
      raise NotImplementedError(city_object.type)
    if not self._buildings:
      logging.warning('impossible to remove city_object, the city is empty\n')
    else:
      if city_object in self._buildings:
        self._buildings.remove(city_object)
        # regenerate hash map
        self._city_objects_dictionary = {}
        self._city_objects_alias_dictionary = {}
        for i, _building in enumerate(self._buildings):
          self._city_objects_dictionary[_building.name] = i
          for alias in _building.aliases:
            if alias in self._city_objects_alias_dictionary:
              self._city_objects_alias_dictionary[alias].append(i)
            else:
              self._city_objects_alias_dictionary[alias] = [i]

  @property
  def srs_name(self) -> Union[None, str]:
    """
    Get city srs name
    :return: None or str
    """
    return self._srs_name

  @name.setter
  def name(self, value):
    """
    Set city name
    :param value:str
    """
    if value is not None:
      self._name = str(value)

  @staticmethod
  def load(city_filename) -> City:
    """
    Load a city saved with city.save(city_filename)
    :param city_filename: city filename
    :return: City
    """
    if sys.platform == 'win32':
      pathlib.PosixPath = pathlib.WindowsPath
    elif sys.platform == 'linux':
      pathlib.WindowsPath = pathlib.PosixPath

    with open(city_filename, 'rb') as file:
      return pickle.load(file)

  def save(self, city_filename):
    """
    Save a city into the given filename
    :param city_filename: destination city filename
    :return: None
    """
    with open(city_filename, 'wb') as file:
      pickle.dump(self, file)

  def save_compressed(self, city_filename):
    """
    Save a city into the given filename
    :param city_filename: destination city filename
    :return: None
    """
    with bz2.BZ2File(city_filename, 'wb') as file:
      pickle.dump(self, file)

  def region(self, center, radius) -> City:
    """
    Get a region from the city
    :param center: specific point in space [x, y, z]
    :param radius: distance to center of the sphere selected in meters
    :return: selected_region_city
    """
    selected_region_lower_corner = [center[0] - radius, center[1] - radius, center[2] - radius]
    selected_region_upper_corner = [center[0] + radius, center[1] + radius, center[2] + radius]
    selected_region_city = City(selected_region_lower_corner, selected_region_upper_corner, srs_name=self.srs_name)
    selected_region_city.climate_file = self.climate_file
    #    selected_region_city.climate_reference_city = self.climate_reference_city
    for city_object in self.city_objects:
      location = city_object.centroid
      if location is not None:
        distance = math.sqrt(math.pow(location[0] - center[0], 2) + math.pow(location[1] - center[1], 2)
                             + math.pow(location[2] - center[2], 2))
        if distance < radius:
          selected_region_city.add_city_object(city_object)
    return selected_region_city

  @property
  def latitude(self) -> Union[None, float]:
    """
    Get city latitude in degrees
    :return: None or float
    """
    return self._latitude

  @latitude.setter
  def latitude(self, value):
    """
    Set city latitude in degrees
    :parameter value: float
    """
    if value is not None:
      self._latitude = float(value)

  @property
  def longitude(self) -> Union[None, float]:
    """
    Get city longitude in degrees
    :return: None or float
    """
    return self._longitude

  @longitude.setter
  def longitude(self, value):
    """
    Set city longitude in degrees
    :parameter value: float
    """
    if value is not None:
      self._longitude = float(value)

  @property
  def time_zone(self) -> Union[None, float]:
    """
    Get city time_zone
    :return: None or float
    """
    return self._time_zone

  @time_zone.setter
  def time_zone(self, value):
    """
    Set city time_zone
    :parameter value: float
    """
    if value is not None:
      self._time_zone = float(value)

  @property
  def buildings_clusters(self) -> Union[List[BuildingsCluster], None]:
    """
    Get buildings clusters belonging to the city
    :return: None or [BuildingsCluster]
    """
    return self._buildings_clusters

  @property
  def parts_consisting_buildings(self) -> Union[List[PartsConsistingBuilding], None]:
    """
    Get parts consisting buildings belonging to the city
    :return: None or [PartsConsistingBuilding]
    """
    return self._parts_consisting_buildings

  @property
  def energy_systems(self) -> Union[List[EnergySystem], None]:
    """
    Get energy systems belonging to the city
    :return: None or [EnergySystem]
    """
    return self._energy_systems

  @property
  def stations(self) -> [Station]:
    """
    Get the sensors stations belonging to the city
    :return: [Station]
    """
    return self._stations

  @property
  def city_objects_clusters(self) -> Union[List[CityObjectsCluster], None]:
    """
    Get city objects clusters belonging to the city
    :return: None or [CityObjectsCluster]
    """
    if self.buildings_clusters is None:
      self._city_objects_clusters = []
    else:
      self._city_objects_clusters = self.buildings_clusters
    if self.parts_consisting_buildings is not None:
      self._city_objects_clusters.append(self.parts_consisting_buildings)
    return self._city_objects_clusters

  def add_city_objects_cluster(self, new_city_objects_cluster):
    """
    Add a CityObject to the city
    :param new_city_objects_cluster:CityObjectsCluster
    :return: None or NotImplementedError
    """
    if new_city_objects_cluster.type == 'buildings':
      if self._buildings_clusters is None:
        self._buildings_clusters = []
      self._buildings_clusters.append(new_city_objects_cluster)
    elif new_city_objects_cluster.type == 'building_parts':
      if self._parts_consisting_buildings is None:
        self._parts_consisting_buildings = []
      self._parts_consisting_buildings.append(new_city_objects_cluster)
    else:
      raise NotImplementedError

  @property
  def copy(self) -> City:
    """
    Get a copy of the current city
    """
    return copy.deepcopy(self)

  def merge(self, city) -> City:
    """
    Return a merged city combining the current city and the given one
    :return: City
    """
    merged_city = self.copy
    for building in city.buildings:
      if merged_city.city_object(building.name) is None:
        # building is new so added to the city
        merged_city.add_city_object(copy.deepcopy(building))
      else:
        # keep the one with less radiation
        parameter_city_building_total_radiation = 0
        for surface in building.surfaces:
          if surface.global_irradiance:
            parameter_city_building_total_radiation += surface.global_irradiance[cte.YEAR].iloc[0, 0]

        merged_city_building_total_radiation = 0
        for surface in merged_city.city_object(building.name).surfaces:
          if surface.global_irradiance:
            merged_city_building_total_radiation += surface.global_irradiance[cte.YEAR].iloc[0, 0]

        if merged_city_building_total_radiation == 0:
          merged_city.remove_city_object(merged_city.city_object(building.name))
          merged_city.add_city_object(building)
        elif merged_city_building_total_radiation > parameter_city_building_total_radiation > 0:
          merged_city.remove_city_object(merged_city.city_object(building.name))
          merged_city.add_city_object(building)

    return merged_city

  @property
  def level_of_detail(self) -> LevelOfDetail:
    """
    Get level of detail of different aspects of the city: geometry, construction and usage
    :return: LevelOfDetail
    """
    return self._level_of_detail

  @property
  def energy_systems_connection_table(self) -> Union[None, DataFrame]:
    """
    Get energy systems connection table which includes at least two columns: energy_system_type and associated_building
    and may also include dimensioned_energy_system and connection_building_to_dimensioned_energy_system
    :return: DataFrame
    """
    return self._energy_systems_connection_table

  @energy_systems_connection_table.setter
  def energy_systems_connection_table(self, value):
    """
    Set energy systems connection table which includes at least two columns: energy_system_type and associated_building
    and may also include dimensioned_energy_system and connection_building_to_dimensioned_energy_system
    :param value: DataFrame
    """
    self._energy_systems_connection_table = value

  @property
  def generic_energy_systems(self) -> dict:
    """
    Get dictionary with generic energy systems installed in the city
    :return: dict
    """
    return self._generic_energy_systems

  @generic_energy_systems.setter
  def generic_energy_systems(self, value):
    """
    Set dictionary with generic energy systems installed in the city
    :return: dict
    """
    self._generic_energy_systems = value



"""
Cost catalog
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2023 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
"""

import xmltodict
from hub.catalog_factories.catalog import Catalog
from hub.catalog_factories.data_models.cost.archetype import Archetype
from hub.catalog_factories.data_models.cost.content import Content
from hub.catalog_factories.data_models.cost.capital_cost import CapitalCost
from hub.catalog_factories.data_models.cost.chapter import Chapter
from hub.catalog_factories.data_models.cost.item_description import ItemDescription
from hub.catalog_factories.data_models.cost.operational_cost import OperationalCost
from hub.catalog_factories.data_models.cost.fuel import Fuel
from hub.catalog_factories.data_models.cost.income import Income


class MontrealCustomCatalog(Catalog):
  """
  Montreal custom catalog class
  """
  def __init__(self, path):
    path = (path / 'montreal_costs.xml').resolve()
    with open(path, 'r', encoding='utf-8') as xml:
      self._archetypes = xmltodict.parse(xml.read(), force_list='archetype')

    # store the full catalog data model in self._content
    self._content = Content(self._load_archetypes())

  @staticmethod
  def _item_with_threesome(entry, item_type):
    _reposition = float(entry[item_type]['reposition']['#text'])
    _reposition_unit = entry[item_type]['reposition']['@cost_unit']
    _investment = float(entry[item_type]['investment_cost']['#text'])
    _investment_unit = entry[item_type]['investment_cost']['@cost_unit']
    _lifetime = float(entry[item_type]['lifetime_equipment']['#text'])
    _item_description = ItemDescription(item_type,
                                        initial_investment=_investment,
                                        initial_investment_unit=_investment_unit,
                                        reposition=_reposition,
                                        reposition_unit=_reposition_unit,
                                        lifetime=_lifetime)
    return _item_description

  @staticmethod
  def _item_with_refurbishment_values(entry, item_type):
    _refurbishment = float(entry[item_type]['refurbishment_cost']['#text'])
    _refurbishment_unit = entry[item_type]['refurbishment_cost']['@cost_unit']
    _item_description = ItemDescription(item_type,
                                        refurbishment=_refurbishment,
                                        refurbishment_unit=_refurbishment_unit)
    return _item_description

  def _get_capital_costs(self, entry):
    general_chapters = []
    shell = entry['B_shell']
    items_list = []
    item_type = 'B10_superstructure'
    item_description = self._item_with_refurbishment_values(shell, item_type)
    items_list.append(item_description)
    for item in shell['B20_envelope']:
      item_type = item
      item_description = self._item_with_refurbishment_values(shell['B20_envelope'], item_type)
      items_list.append(item_description)
    item_type = 'B3010_opaque_roof'
    item_description = self._item_with_refurbishment_values(shell['B30_roofing'], item_type)
    items_list.append(item_description)
    general_chapters.append(Chapter('B_shell', items_list))
    items_list = []
    item_type = 'D301010_photovoltaic_system'
    services = entry['D_services']
    item_description = self._item_with_threesome(services['D30_hvac']['D3010_energy_supply'], item_type)
    items_list.append(item_description)
    item_type_list = ['D3020_heat_generating_systems', 'D3030_cooling_generation_systems', 'D3040_distribution_systems',
                      'D3080_other_hvac_ahu']
    for item_type in item_type_list:
      item_description = self._item_with_threesome(services['D30_hvac'], item_type)
      items_list.append(item_description)
    item_type = 'D5020_lighting_and_branch_wiring'
    item_description = self._item_with_threesome(services['D50_electrical'], item_type)
    items_list.append(item_description)
    general_chapters.append(Chapter('D_services', items_list))
    allowances = entry['Z_allowances_overhead_profit']
    design_allowance = float(allowances['Z10_design_allowance']['#text']) / 100
    overhead_and_profit = float(allowances['Z20_overhead_profit']['#text']) / 100
    _capital_cost = CapitalCost(general_chapters, design_allowance, overhead_and_profit)

    return _capital_cost

  @staticmethod
  def _get_operational_costs(entry):
    fuels = []
    for item in entry['fuels']['fuel']:
      fuel_type = item['@fuel_type']
      fuel_variable = float(item['variable']['#text'])
      fuel_variable_units = item['variable']['@cost_unit']
      fuel_fixed_monthly = None
      fuel_fixed_peak = None
      if fuel_type == 'electricity':
        fuel_fixed_monthly = float(item['fixed_monthly']['#text'])
        fuel_fixed_peak = float(item['fixed_power']['#text']) / 1000
      elif fuel_type == 'gas':
        fuel_fixed_monthly = float(item['fixed_monthly']['#text'])
      fuel = Fuel(fuel_type,
                  fixed_monthly=fuel_fixed_monthly,
                  fixed_power=fuel_fixed_peak,
                  variable=fuel_variable,
                  variable_units=fuel_variable_units)
      fuels.append(fuel)
    heating_equipment_maintenance = float(entry['maintenance']['heating_equipment']['#text']) / 1000
    cooling_equipment_maintenance = float(entry['maintenance']['cooling_equipment']['#text']) / 1000
    photovoltaic_system_maintenance = float(entry['maintenance']['photovoltaic_system']['#text'])
    co2_emissions = float(entry['co2_cost']['#text'])
    _operational_cost = OperationalCost(fuels,
                                        heating_equipment_maintenance,
                                        cooling_equipment_maintenance,
                                        photovoltaic_system_maintenance,
                                        co2_emissions)
    return _operational_cost

  def _load_archetypes(self):
    _catalog_archetypes = []
    archetypes = self._archetypes['archetypes']['archetype']
    for archetype in archetypes:
      function = archetype['@function']
      municipality = archetype['@municipality']
      country = archetype['@country']
      lod = float(archetype['@lod'])
      currency = archetype['currency']
      capital_cost = self._get_capital_costs(archetype['capital_cost'])
      operational_cost = self._get_operational_costs(archetype['operational_cost'])
      end_of_life_cost = float(archetype['end_of_life_cost']['#text'])
      construction = float(archetype['incomes']['subsidies']['construction']['#text'])
      hvac = float(archetype['incomes']['subsidies']['hvac']['#text'])
      photovoltaic_system = float(archetype['incomes']['subsidies']['photovoltaic']['#text'])
      electricity_exports = float(archetype['incomes']['electricity_export']['#text']) / 1000 / 3600
      reduction_tax = float(archetype['incomes']['tax_reduction']['#text']) / 100
      income = Income(construction_subsidy=construction,
                      hvac_subsidy=hvac,
                      photovoltaic_subsidy=photovoltaic_system,
                      electricity_export=electricity_exports,
                      reductions_tax=reduction_tax)
      _catalog_archetypes.append(Archetype(lod,
                                           function,
                                           municipality,
                                           country,
                                           currency,
                                           capital_cost,
                                           operational_cost,
                                           end_of_life_cost,
                                           income))
    return _catalog_archetypes

  def names(self, category=None):
    """
    Get the catalog elements names
    :parm: for costs catalog category filter does nothing as there is only one category (archetypes)
    """
    _names = {'archetypes': []}
    for archetype in self._content.archetypes:
      _names['archetypes'].append(archetype.name)
    return _names

  def entries(self, category=None):
    """
    Get the catalog elements
    :parm: for costs catalog category filter does nothing as there is only one category (archetypes)
    """
    return self._content

  def get_entry(self, name):
    """
    Get one catalog element by names
    :parm: entry name
    """
    for entry in self._content.archetypes:
      if entry.name.lower() == name.lower():
        return entry
    raise IndexError(f"{name} doesn't exists in the catalog")



"""
Join module
SPDX - License - Identifier: LGPL - 3.0 - or -later
Copyright © 2022 Concordia CERC group
Project Coder Pilar Monsalvete Alvarez de Uribarri pilar.monsalvete@concordia.ca
Code contributors: Guille guille.gutierrezmorote@concordia.ca
"""

from hub.city_model_structure.transport.traffic_node import TrafficNode


class Join(TrafficNode):
  """
  Join class
  """

  def __init__(self, name, coordinates, nodes):
    self._nodes = nodes
    edges = []
    prohibitions = []
    connections = []
    for node in self._nodes:
      edges = edges + node.edges
      prohibitions = prohibitions + node.prohibitions
      connections = connections + node.connections
    super().__init__(name, coordinates, edges=edges, prohibitions=prohibitions, connections=connections,
                     node_type='Join')
